# -*- coding: utf-8 -*-
# !/usr/bin/python

""" Tkinter interface made for Live Coding with Python syntax highlighting """

# This removed blurry fonts on Windows
try:
    from ctypes import windll
    windll.shcore.SetProcessDpiAwareness(1)
except Exception:
    pass
# Tkinter Interface
from FoxDotEditor.tkimport import *
# Custom app modules
from renardo_lib.Code import write_to_file
from renardo_lib.Utils import get_pypi_version
from .Format import *
from .AppFunctions import *
from .Console import console
from .Prompt import TextPrompt
from .BracketHandler import BracketHandler
from .TextBox import ThreadedText
from .LineNumbers import LineNumbers
from .MenuBar import MenuBar, PopupMenu
from .Treeview import TreeView
from .SampleChart import SampleChart
from .SearchBar import SearchBar
from functools import partial
# from distutils.version import LooseVersion as VersionNumber
import webbrowser
import os
import re
import socket
# Code execution
from renardo_lib.Settings import *
from renardo_lib.Code import execute
from renardo_lib.ServerManager import TempoServer


# App object
class workspace:
    default_font = FONT
    namespace = {}

    def __init__(self, CodeClass):
        # Configure Renardo's namespace to include the editor
        CodeClass.namespace['GUI'] = self
        CodeClass.namespace['Player'].widget = self
        self.version = this_version = '0.9.13'
        pypi_version = get_pypi_version()
        self.theme = COLOR_THEME

        def check_versions():
            # if pypi_version is not None and VersionNumber(pypi_version)
            # > VersionNumber(this_version):
            # tkMessageBox.showinfo("New version available", "There is a
            # new version of Renardo available from PyPI. Upgrade by going
            # to your command prompt and running:\n\n
            # pip install Renardo --upgrade")
            return
        # Used for docstring prompt
        self.namespace = CodeClass.namespace
        # Set up master widget
        self.root = tb.Window(themename=self.theme)
        self.root.title("FoxDot >> Renardo")
        self.root.minsize(800, 600)
        self.width = 1024
        self.height = 576
        self.root.geometry(f"{self.width}x{self.height}")
        self.root.rowconfigure(0, weight=1)  # Text box
        self.root.rowconfigure(1, weight=0)  # Search box
        self.root.rowconfigure(2, weight=0)  # Separator
        self.root.rowconfigure(3, weight=0)  # Console
        # self.root.rowconfigure(3, weight=0)  # Console
        self.root.grid_columnconfigure(0, weight=0)  # tree view
        self.root.grid_columnconfigure(1, weight=0)  # line numbers
        self.root.grid_columnconfigure(2, weight=1)  # Scrollbars
        self.root.protocol("WM_DELETE_WINDOW", self.kill)
        # Track whether user wants transparent background
        self.transparent = BooleanVar()
        self.transparent.set(False)
        self.using_alpha = USE_ALPHA
        # Boolean for connection
        self.listening_for_connections = BooleanVar()
        self.listening_for_connections.set(False)
        # Boolean for views
        self.fullscreen_toggled = BooleanVar()
        self.fullscreen_toggled.set(False)
        self.menu_toggled = BooleanVar()
        self.menu_toggled.set(MENU_ON_STARTUP)
        self.treeview_toggled = BooleanVar()
        self.treeview_toggled.set(TREEVIEW_ON_STARTUP)
        self.linenumbers_toggled = BooleanVar()
        self.linenumbers_toggled.set(LINENUMBERS_ON_STARTUP)
        self.searchbar_toggled = BooleanVar()
        self.searchbar_toggled.set(False)
        self.console_toggled = BooleanVar()
        self.console_toggled.set(CONSOLE_ON_STARTUP)
        # Boolean for beat counter
        self.show_counter = BooleanVar()
        self.show_counter.set(False)
        # Boolean for showing auto-complete prompt
        self.show_prompt = BooleanVar()
        self.show_prompt.set(True)
        # --- Set icon
        try:
            # Use .ico file by default
            self.root.iconbitmap(FOXDOT_ICON)
        except TclError:
            # Use .gif if necessary
            self.root.tk.call('wm',
                              'iconphoto',
                              self.root._w,
                              PhotoImage(file=FOXDOT_ICON_GIF))
        # --- Setup font
        system_fonts = tkFont.families()
        self.codefont = "CodeFont"  # name for font
        if self.default_font not in system_fonts:
            if SYSTEM == WINDOWS and "Consolas" in system_fonts:
                self.default_font = "Consolas"
            elif SYSTEM == MAC_OS and "Monaco" in system_fonts:
                self.default_font = "Monaco"
            elif "Ubuntu Mono" in system_fonts:
                self.default_font = "Ubuntu Mono"
            elif "Courier New" in system_fonts:
                self.default_font = "Courier New"
            else:
                self.console_font = self.codefont = self.default_font = "TkFixedFont"
        if self.codefont == "CodeFont":
            self.font = tkFont.Font(font=(self.default_font, 12),
                                    name=self.codefont)
            self.font.configure(family=self.default_font)
            self.console_font = (self.default_font, 12)
        self.help_key = "K" if SYSTEM == MAC_OS else "H"
        # --- start create menu
        self.menu = MenuBar(self, visible=MENU_ON_STARTUP)
        self.popup = PopupMenu(self)
        # Create y-axis scrollbar
        self.y_scroll = tb.Scrollbar(self.root)
        self.y_scroll.grid(row=0, column=3, sticky='ns')
        # Create text box for code
        self.text = ThreadedText(self.root,
                                 padx=5,
                                 pady=5,
                                 bg=colour_map['background'],
                                 fg=colour_map['plaintext'],
                                 insertbackground="white",
                                 font=self.codefont,
                                 yscrollcommand=self.y_scroll.set,
                                 width=100,
                                 height=20,
                                 bd=0,
                                 undo=True,
                                 autoseparators=True,
                                 maxundo=50)
        self.text.grid(row=0, column=2, sticky="nsew")
        self.y_scroll.config(command=self.text.yview)
        self.text.focus_set()
        self.text_as_string = ""
        # Create box for line numbers
        self.linenumbers = LineNumbers(self, width=50,
                                       bg=colour_map['background'],
                                       bd=0,
                                       highlightthickness=0)
        self.linenumbers.grid(row=0, column=1, sticky='nsew')
        # Docstring prompt label
        self.prompt = TextPrompt(self)
        # Key bindings (Use command key on Mac)
        ctrl = "Command" if SYSTEM == MAC_OS else "Control"
        alt = "Option" if SYSTEM == MAC_OS else "Alt"
        self.text.bind("<<Modified>>", self.on_text_modified)
        self.text.bind("<Return>", self.newline)
        self.text.bind("<BackSpace>", self.backspace)
        self.text.bind("<Delete>", self.delete)
        self.text.bind("<Tab>", self.tab)
        self.text.bind("<Escape>", self.toggle_fullscreen)
        # self.text.bind("<F11>", self.toggle_true_fullscreen)
        self.text.bind("<Key>", self.keypress)
        self.text.bind("<Button-{}>".format(2 if SYSTEM == MAC_OS else 3),
                       self.show_popup)
        self.text.bind("<{}-BackSpace>".format(ctrl), self.delete_word)
        self.text.bind("<{}-Delete>".format(ctrl), self.delete_next_word)
        self.text.bind("<{}-Return>".format(ctrl), self.exec_block)
        self.text.bind("<{}-Return>".format(alt), self.exec_line)
        # Directional movement
        self.text.bind("<Up>", self.key_up)
        self.text.bind("<Down>", self.key_down)
        self.text.bind("<{}-Left>".format(ctrl), self.move_word_left)
        self.text.bind("<{}-Right>".format(ctrl), self.move_word_right)
        self.text.bind("<{}-Shift-Right>".format(ctrl), self.select_word_right)
        self.text.bind("<{}-Shift-Left>".format(ctrl), self.select_word_left)
        self.text.bind("<Alt_L>", lambda event: "break")
        # Fixes a MacOS tilde bug.
        self.text.bind("<Alt-n>", lambda event: "break")
        self.text.bind("<{}-a>".format(ctrl), self.select_all)
        self.text.bind("<{}-d>".format(ctrl), self.duplicate_line)
        self.text.bind("<{}-period>".format(ctrl), self.killall)
        self.text.bind("<Alt-period>", self.releaseNodes)
        self.text.bind("<{}-c>".format(ctrl), self.edit_copy)
        self.text.bind("<{}-x>".format(ctrl), self.edit_cut)
        self.text.bind("<{}-v>".format(ctrl), self.edit_paste)
        self.text.bind("<{}-bracketright>".format(ctrl), self.indent)
        self.text.bind("<{}-bracketleft>".format(ctrl), self.unindent)
        self.text.bind("<{}-equal>".format(ctrl), self.zoom_in)
        self.text.bind("<{}-minus>".format(ctrl), self.zoom_out)
        self.text.bind("<{}-z>".format(ctrl), self.undo)
        self.text.bind("<{}-y>".format(ctrl), self.redo)
        self.text.bind("<{}-s>".format(ctrl), self.save)
        self.text.bind("<{}-o>".format(ctrl), self.openfile)
        self.text.bind("<{}-n>".format(ctrl), self.newfile)
        self.text.bind("<{}-m>".format(ctrl), self.toggle_menu)
        self.text.bind("<{}-u>".format(ctrl), self.toggle_treeview)
        self.text.bind("<{}-f>".format(ctrl), self.toggle_searchbar)
        self.text.bind("<{}-k>".format(ctrl), self.toggle_console)
        self.text.bind("<{}-0>".format(ctrl), self.toggle_linenumbers)
        self.text.bind("<{}-p>".format(ctrl), self.open_preferences)
        # insert lambda
        self.text.bind("<{}-l>".format(ctrl),
                       lambda event: self.insert_char(u"\u03BB"))
        self.text.bind("<{}-t>".format(ctrl),
                       lambda event: self.insert_char("~"))
        self.text.bind("<{}-{}>".format(ctrl, self.help_key.lower()),
                       self.help)
        # Number pad
        for event in ["KP_Right", "KP_Left", "KP_Up", "KP_Down", "KP_Delete",
                      "KP_Home",  "KP_End",  "KP_Next", "KP_Prior"]:
            try:
                event1 = "<{}>".format(event)
                event2 = "<{}-{}>".format(ctrl, event)
                event3 = "<{}-{}>".format("Shift", event)
                event4 = "<{}-{}-{}>".format("Shift", ctrl, event)
                event5 = "<{}-{}-{}>".format(ctrl, "Shift", event)
                self.text.bind(event1, partial(lambda *args: self.text.event_generate(args[0]), event1.replace("KP_", "")))
                self.text.bind(event2, partial(lambda *args: self.text.event_generate(args[0]), event2.replace("KP_", "")))
                self.text.bind(event3, partial(lambda *args: self.text.event_generate(args[0]), event3.replace("KP_", "")))
                self.text.bind(event4, partial(lambda *args: self.text.event_generate(args[0]), event4.replace("KP_", "")))
                self.text.bind(event5, partial(lambda *args: self.text.event_generate(args[0]), event5.replace("KP_", "")))
            except TclError:
                pass
        try:
            self.text.bind("<KP_Enter>", self.newline)
        except TclError:
            pass
        # Toggle console button keybind
        try:
            self.text.bind("<{}-#>".format(ctrl), self.toggle_console)
            self.toggle_key = "#"
        except Exception:
            self.text.bind("<{}-K>".format(ctrl), self.toggle_console)
            self.toggle_key = "K"
        # Save feature variabes
        self.saved = False
        self.file = None
        self.filename = None
        # --- define bracket behaviour
        self.bracketHandler = BracketHandler(self)
        # Set tag names and config for specific colours
        for tier in tag_weights:
            for tag_name in tier:
                self.text.tag_config(tag_name, foreground=colour_map[tag_name])
        # Create searchbar
        self.searchbar = SearchBar(self)
        if self.searchbar_toggled.get() is True:
            self.searchbar.show()
        # Create treeview
        self.treeview = TreeView(self)
        if self.treeview_toggled.get() is True:
            self.treeview.show()
        else:
            self.treeview.hide()
        if self.linenumbers_toggled.get() is True:
            self.linenumbers.show()
        else:
            self.linenumbers.hide()
        self.root.bind("<Button-1>", self.mouse_press)
        # Store original location of cursor
        self.origin = "origin"
        self.text.mark_set(self.origin, INSERT)
        self.text.mark_gravity(self.origin, LEFT)
        # --- Create console
        self.console = console(self)
        sys.stdout = self.console
        # Ask after widget loaded
        self.linenumbers.redraw()  # ToDo: move to generic redraw functions
        # Check temporary file
        def recover_work():
            with open(FOXDOT_TEMP_FILE) as f:
                text = f.read()
            if len(text):
                loading = tkMessageBox.askyesno("Load unsaved work?", "Your code wasn't saved last time you used Renardo, do you want to load any unsaved work?")
                self.root.update()
                if loading:
                    self.set_all(text)
                else:
                    self.clear_temp_file()
                self.text_as_string = self.get_all()
            # Execute startup file
            return execute.load_startup_file()
        # Check online if a new version if available
        if CHECK_FOR_UPDATE:
            self.root.after(90, check_versions)
        # Ask after widget loaded
        if RECOVER_WORK:
            self.root.after(100, recover_work)
        # Check transparency on startup
        if TRANSPARENT_ON_STARTUP:
            self.transparent.set(True)
            self.root.after(100, self.toggle_transparency)

    def set_window_title(self, text="Live Coding with Python & SuperCollider"):
        return self.root.title("Renardo v{} - {}".format(self.version, text))

    def run(self):
        """ Starts the Tk mainloop for the master widget """
        while True:
            try:
                self.root.mainloop()
                break
            # Temporary fix to unicode issues with Mac OS
            except (UnicodeDecodeError):
                pass
            except (KeyboardInterrupt, SystemExit):
                # Clean exit
                self.terminate()
                break
        # If the work has not been saved, store in a temporary file
        if not self.saved:
            self.set_temp_file(self.text_as_string)
        return

    def toggle_fullscreen(self, event=None, zoom=False):
        """ Zoom the screen - close with Escape """
        if self.root.attributes('-fullscreen'):
            self.root.attributes('-fullscreen', 0)
            self.fullscreen_toggled.set(False)
        elif zoom:
            self.root.attributes('-fullscreen', 1)
            self.fullscreen_toggled.set(True)
        return

    def reload(self):
        """ Reloads synths / samples """
        return self.namespace['_reload_synths'].__call__()

    def read(self):
        return self.text.get("1.0", END)

    def keypress(self, event=None):
        """ Handles any keypress """
        # For non string characters, return normally
        self.text.tag_delete("tag_open_brackets")
        if not event.char or isHex(event.char):
            self.inbrackets = False
            self.update_prompt(visible=False)
            return
        # Add character to text box
        else:
            self.delete_selection()
            index = self.text.index(INSERT)
            self.text.insert(index, event.char)  # should modified be called?
            self.text.edit_separator()
            self.update(event)
            # File is unsaved
            self.saved = False
            self.text_as_string = self.get_all()
        return "break"

    """
        Getting blocks / lines
    """
    def exec_line(self, event=None, insert=INSERT):
        """ Highlights a single line and executes """
        line, column = index(self.text.index(insert))
        a, b = "%d.0" % line, "%d.end" % line
        self.highlight(a, b, "red")
        try:
            execute(self.text.get(a, b))
            # execute.update_line_numbers(self.text, a, b)
        except Exception:
            pass
        self.root.after(200, self.unhighlight)
        return "break"

    def exec_block(self, event=None, insert=INSERT):
        """ Method to highlight block of code and execute """
        try:
            # Evaluate selection
            a = index(self.text.index(SEL_FIRST))[0]
            b = index(self.text.index(SEL_LAST))[0] + 1
        except TclError:
            # Get start and end of the buffer
            start, end = "1.0", self.text.index(END)
            lastline = int(end.split('.')[0]) + 1
            # Indicies of block to execute
            block = [0, 0]
            # 1. Get position of cursor
            cur_x, cur_y = index(self.text.index(insert))
            # 2. Go through line by line (back) and see what it's value is
            for line in range(cur_x, 0, -1):
                if not self.text.get("%d.0" % line, "%d.end" % line).strip():
                    break
            block[0] = line
            # 3. Iterate forwards until we get two \n\n or index==END
            for line in range(cur_x, lastline):
                if not self.text.get("%d.0" % line, "%d.end" % line).strip():
                    break
            block[1] = line
            # Now we have the lines of code!
            a, b = block
        if a == b:
            b += 1
        for line in range(a, b):
            start = "%d.0" % line
            end = "%d.end" % line
            # Highlight text only to last character, not whole line
            if len(self.text.get(start, end).strip()) > 0:
                self.highlight(start, end)
        # Convert line numbers to Tkinter indices
        a, b = ("%d.0" % n for n in (a, b))
        # Execute the python code
        try:
            execute(self.text.get(a, b))
            # execute.update_line_numbers(self.text, a, b)
        except Exception:
            pass
        # Unhighlight the line of text
        self.root.after(200, self.unhighlight)
        return "break"

    # Scheduling tasks
    # ----------------
    def addTask(self, target, args=(), kwargs={}):
        self.text.queue.put((target, args, kwargs))
        return

    def insert_char(self, char):
        """ Inserts a character into the text editor at the INSERT cursor
            then updates the syntax highlighting etc """
        self.text.insert(INSERT, char)
        self.update()
        return "break"

    def insert_lambda_symbol(self, event):
        return self.insert_char(u"\u03BB")

    # Undo action: Ctrl+Z
    # --------------------
    def undo(self, event=None):
        try:
            self.text.edit_undo()
            # self.update_all()
        except TclError:
            pass
        return "break"

    def redo(self, event=None):
        try:
            self.text.edit_redo()
            # self.update_all()
        except TclError:
            pass
        return "break"

    # Help feature: Ctrl+H
    # ---------------------
    def help(self, event=None):
        if SYSTEM == MAC_OS:
            ctrl = "Cmd"
        else:
            ctrl = "Ctrl"
        print("Renardo Help:")
        print("-----------------------------------------")
        print("{}+Return: Execute code".format(ctrl))
        print("{}+.     : Stop all sound".format(ctrl))
        print("{}+=     : Increase font size".format(ctrl))
        print("{}+-     : Decrease font size".format(ctrl))
        print("{}+L     : Insert lambda symbol".format(ctrl))
        print("{}+T     : Insert tilde symbol".format(ctrl))
        print("{}+S     : Save your work".format(ctrl))
        print("{}+O     : Open a file".format(ctrl))
        print("{}+M     : Toggle Menu".format(ctrl))
        print("{}+K     : Toggle Console".format(ctrl))
        print("{}+U     : Toggle Treeview".format(ctrl))
        print("{}+F     : Toggle Searchbar".format(ctrl))
        print("{}+P     : Open Preferences".format(ctrl))
        print("print(SynthDefs)      : View available SynthDefs")
        print("print(Samples)        : View character-to-sample mapping")
        print("print(FxList)         : View audio effects")
        print("print(Attributes)     : View Player attributes")
        print("print(PatternMethods) : View Pattern methods")
        print("----------------------------------------------------------")
        print("Please visit foxdot.org / renardo.org for more information")
        print("----------------------------------------------------------")
        return "break"

    # Save the current text: Ctrl+s
    # ------------------------------
    def save(self, event=None):
        """ Saves the contents of the text editor """
        text = self.text.get("0.0", END)
        if not self.saved:
            self.filename = tkFileDialog.asksaveasfilename(filetypes=[("Python files", ".py")],
                                                           defaultextension=".py")
        if self.filename:
            write_to_file(self.filename, text)
            self.saved = True
            print("Saved '{}'".format(self.filename))
            # Remove tmp file
            self.clear_temp_file()
        return bool(self.filename)

    # Open save
    def saveAs(self, event=None):
        text = self.text.get("0.0", END)
        self.filename = tkFileDialog.asksaveasfilename(filetypes=[("Python files", ".py")],
                                                       defaultextension=".py")
        if not self.filename:
            pass
        else:
            write_to_file(self.filename, text)
            self.saved = True
            # Remove tmp file
            self.clear_temp_file()
        if self.saved:
            print("Save successful!")
        return bool(self.filename)

    # Open a file: Ctrl+o
    # --------------------
    def openfile(self, event=None):
        path = tkFileDialog.askopenfilename()
        if path != "":
            try:
                f = open(path)
                text = f.read()
                f.close()
                self.set_all(text)
                self.set_window_title(path)
            except Exception:
                pass
        return "break"

    def opentvfile(self, file):
        f = open(file)
        text = f.read()
        f.close()
        self.set_all(text)
        self.set_window_title(file)

    def loadfile(self, path):
        try:
            f = open(path, encoding="utf8")
        except Exception as e:
            return print("{} error occurred when loading file:\n    - '{}'".format(e.__class__.__name__, path))
        self.set_all(f.read())
        f.close()
        return

    def newfile(self, event=None):
        ''' Clears the document and asks if the user wants to save '''
        answer = tkMessageBox.askyesnocancel("New file", "Save your work before creating a new document?")
        if answer is not None:
            if answer is True:
                if not self.save():
                    return "break"
            self.saved = False
            self.filename = ''
            self.set_all("")
            self.set_window_title()
        return "break"

    def export_console(self):
        fn = tkFileDialog.asksaveasfilename(filetypes=[("Plain Text File", ".txt")],
                                            defaultextension=".txt")
        with open(fn, 'w') as f:
            f.write(self.console.read())
        return

    def open_preferences(self, event=None):
        from FoxDotEditor.Preferences import Preferences
        Preferences().start()
        return

    def open_samples_folder(self):
        import subprocess
        if SYSTEM == WINDOWS:
            cmd = 'explorer'
        elif SYSTEM == MAC_OS:
            cmd = 'open'
        else:
            cmd = 'xdg-open'
        try:
            subprocess.Popen([cmd, FOXDOT_SND])
        except OSError as e:
            print(e)
            print("Hmm... Looks like we couldn't open the directory but you can find the samples in {}".format(FOXDOT_SND))
        return

    def open_samples_chart_app(self):
        try:
            sample_chart = SampleChart()
        except OSError as e:
            print(e)
            print("Hmm... Looks like we couldn't open the chart app.")
        return

    # Toggle linenumbers: ctrl+0
    # -----------------------------
    def toggle_linenumbers(self, event=None):
        linenumbers_toggle = self.linenumbers_toggled.get()
        if linenumbers_toggle:
            self.menu.viewmenu.entryconfigure(1, label="Show Line Numbers")
            self.linenumbers.hide()
            # self.text.config(height=self.text.cget('height')+self.searchbar.height)
            self.linenumbers_toggled.set(False)
        elif not linenumbers_toggle:
            self.menu.viewmenu.entryconfigure(1, label="Hide Line Numbers")
            self.linenumbers.show()
            # self.text.config(height=self.text.cget('height')-self.searchbar.height)
            self.linenumbers_toggled.set(True)
        return

    # Toggle console: Ctrl+#
    # -----------------------------
    def toggle_console(self, event=None):
        console_toggle = self.console_toggled.get()
        if console_toggle:
            self.menu.viewmenu.entryconfigure(5, label="Show Console")
            self.console.hide()
            self.text.config(height=self.text.cget('height')+self.console.height)
            self.console_toggled.set(False)
        elif not console_toggle:
            self.menu.viewmenu.entryconfigure(5, label="Hide Console")
            self.console.show()
            self.text.config(height=self.text.cget('height')-self.console.height)
            self.console_toggled.set(True)
        return

    # Toggle treeview: ctrl+t
    # -----------------------------
    def toggle_treeview(self, event=None):
        treeview_toggle = self.treeview_toggled.get()
        if treeview_toggle:
            self.menu.viewmenu.entryconfigure(2, label="Show Treeview")
            self.treeview.hide()
            # self.text.config(height=self.text.cget('height')+self.searchbar.height)
            self.treeview_toggled.set(False)
        elif not treeview_toggle:
            self.menu.viewmenu.entryconfigure(2, label="Hide Treeview")
            self.treeview.show()
            # self.text.config(height=self.text.cget('height')-self.searchbar.height)
            self.treeview_toggled.set(True)
        return

    # Toggle searchbar: ctrl+f
    # -----------------------------
    def toggle_searchbar(self, event=None):
        searchbar_toggle = self.searchbar_toggled.get()
        if searchbar_toggle:
            self.menu.viewmenu.entryconfigure(3, label="Show Searchbar")
            self.searchbar.hide()
            # self.text.config(height=self.text.cget('height')+self.searchbar.height)
            self.searchbar_toggled.set(False)
        elif not searchbar_toggle:
            self.menu.viewmenu.entryconfigure(3, label="Hide Searchbar")
            self.searchbar.show()
            # self.text.config(height=self.text.cget('height')-self.searchbar.height)
            self.searchbar_toggled.set(True)
        return

    def toggle_menu(self, event=None):
        self.menu.toggle()
        if self.menu_toggled.get() is True:
            self.menu.viewmenu.entryconfigure(0, label="Show Menu")
            self.menu_toggled.set(False)
        elif self.menu_toggled.get() is False:
            self.menu.viewmenu.entryconfigure(0, label="Hide Menu")
            self.menu_toggled.set(True)
        return "break"

    def toggle_sc3_plugins(self, event=None):
        """
        Allows you to change the SC3 plugins variable from the editor.
        Restart of the editor is required.
        """
        with open(FOXDOT_CONFIG_FILE, "r") as f:
            lines = f.readlines()
        with open(FOXDOT_CONFIG_FILE, "w") as f:
            for line in lines:
                if "SC3_PLUGINS" in line:
                    f.write("SC3_PLUGINS={}\n".format(not bool(SC3_PLUGINS)))
                else:
                    f.write(line)
        # Pop-up to tell the user a restart is required
        tkMessageBox.showwarning(title="Just a heads up",
                                 message="Restart of Renardo is required for the changes to take effect")
        return "break"

    def toggle_transparency(self, event=None):
        """
        Sets the text and console background to black and then removes all
        black pixels from the GUI
        """
        setting_transparent = self.transparent.get()
        try:
            if setting_transparent:
                if not self.using_alpha:
                    try:
                        alpha = "#000001" if SYSTEM == WINDOWS else "systemTransparent"
                        # self.text.config(background=alpha)
                        self.linenumbers.config(background=alpha)
                        self.console.config(background=alpha)

                        if SYSTEM == WINDOWS:
                            self.root.wm_attributes('-transparentcolor', alpha)
                        else:
                            self.root.wm_attributes("-transdef toggle_prparent", True)
                    except TclError:
                        self.using_alpha = True
                if self.using_alpha:
                    self.root.wm_attributes("-alpha", ALPHA_VALUE)
            # Re-set the colours
            elif not self.using_alpha:
                # self.text.config(background=colour_map['background'])
                self.linenumbers.config(background=colour_map['background'])
                self.console.config(background="Black")

                if SYSTEM == WINDOWS:
                    self.root.wm_attributes('-transparentcolor', "")
                else:
                    self.root.wm_attributes("-transparent", False)
            else:
                self.root.wm_attributes("-alpha", 1)
        except TclError as e:
            print(e)
        return

    def toggle_prompt(self, event=None):
        self.prompt.toggle()
        return "break"

    def toggle_counter(self, event=None):
        self.console.counter.toggle()
        return "break"

    # Copy/paste etc
    def edit_paste(self, event=None):
        """ Pastes any text and updates the IDE """
        self.text.event_generate("<<Paste>>")
        self.update_all()
        return "break"

    def edit_cut(self, event=None):
        self.text.event_generate("<<Cut>>")
        return "break"

    def edit_copy(self, event=None):
        self.text.event_generate("<<Copy>>")
        return "break"

    # Newline
    # --------
    def newline(self, event=None, insert=INSERT):
        """ Adds whitespace to newlines where necessary """
        # Enter from auto prompt
        if self.prompt.visible:
            self.prompt.autocomplete()
            return "break"
        # Remove any highlighted text
        self.delete_selection()
        self.text.tag_delete("tag_open_brackets")
        # Get the text from this line
        i, j = index(self.text.index(insert))
        line = self.text.get("%d.0" % i, "%d.end" % i)
        # Add newline
        self.text.insert(index(i, j), "\n")
        # Update player line numbers
        # execute.update_line_numbers(self.text)
        whitespace = get_tabspace(line)  # amount of whitespace to add
        # If the line was empty, dont add whitespace
        if line.strip() == "":
            whitespace = whitespace[:-tabsize]
        # If the index was in brackets, add one tab size
        elif in_brackets(j, line) or at_function(j, line):
            whitespace = whitespace + self.tabspace()
        # Add the necessary whitespace
        self.text.insert(index(i+1, 0), whitespace)
        # Update the IDE colours
        self.update(event)
        return "break"

    # Tab
    # ----
    def tab(self, event=None, insert=INSERT):
        """ Move selected text forward 4 spaces """
        try:  # Move any selected lines forwards
            a, b = (index(a)[0] for a in (self.text.index(SEL_FIRST),
                                          self.text.index(SEL_LAST)))
            if a < b:
                self.indent(event)
                return "break"
            else:
                self.delete(event)
        except Exception:
            pass
        # Insert white space
        line, column = index(self.text.index(insert))
        self.text.insert(index(line, column), self.tabspace())
        # Update IDE
        self.update(event)
        return "break"

    # Indent: Ctrl+]
    # ---------------
    def indent(self, event=None, insert=INSERT):
        """ Indent the current line or selected text """
        try:
            if not self.text_selected():
                a = index(self.text.index(INSERT))
                a = a[0], a[1]-1
                b = a[0], a[1]+1
                self.text.tag_add(SEL, index(*a), index(*b))
            sel_a = index(index(self.text.index(SEL_FIRST))[0], 0)
            sel_b = index(index(self.text.index(SEL_LAST))[0], 'end')
            start, end = (index(a) for a in (self.text.index(SEL_FIRST),
                                             self.text.index(SEL_LAST)))
            for row in range(start[0], end[0]+1):
                # Add intentation
                self.text.insert(index(row, 0), self.tabspace())
            self.text.tag_add(SEL, sel_a, sel_b)
        except Exception:
            pass
        return "break"

    # Un-inden: Ctrl+[
    # -----------------
    def unindent(self, event):
        """ Moves the current row or selected text back by 4 spaces """
        if not self.text_selected():
            a = index(self.text.index(INSERT))
            a = a[0], a[1]-1
            b = a[0], a[1]+1
            self.text.tag_add(SEL, index(*a), index(*b))
        sel_a = index(index(self.text.index(SEL_FIRST))[0], 0)
        sel_b = index(index(self.text.index(SEL_LAST))[0], 'end')
        start, end = (index(a) for a in (self.text.index(SEL_FIRST), self.text.index(SEL_LAST)))
        for row in range(start[0], end[0]+1):
            # Unindent
            line = self.text.get(index(row, 0), index(row, 'end'))
            for n, char in enumerate(line[:tabsize]):
                if char != " ":
                    break
            if n > 0:
                self.text.delete(index(row, 0), index(row, n+1))
        self.text.tag_add(SEL, sel_a, sel_b)
        return "break"

    # Deletion
    # ---------
    def backspace(self, event=None, insert=INSERT):
        """ Deletes a character or selected area """
        # If there is a selected area, delete that
        if self.delete_selection():
            # Update player line numbers
            # execute.update_line_numbers(self.text)
            return "break"
        # Handle delete in brackets
        if self.bracketHandler.delete():
            return "break"
        # Else, work out if there is a tab to delete
        line, column = index(self.text.index(insert))
        # If we are at the start of a line, delete that
        if column == 0:
            self.update(event)
            # Update player line numbers
            # execute.update_line_numbers(self.text, start="%d.0" % (line-1), remove=int(line!=1))
            self.text.delete(index(line-1, END), insert)
        else:
            tab = index(line, column-tabsize)
            # Check if there's a tab
            if self.text.get(tab, insert) == self.tabspace():
                self.text.delete(tab, insert)
            else:
                self.text.delete(index(line, column-1), insert)
        # Update the IDE
        self.update(event)
        return "break"

    def delete(self, event=None, insert=INSERT):
        """ Delete the next character """
        if not self.delete_selection():
            self.text.delete(self.text.index(insert))
        self.update(event)
        # execute.update_line_numbers(self.text)
        return "break"

    def look(self, direction=-1):
        """ Finds the start of the next / previous word """
        num_words = abs(direction)
        direction = 1 if direction > 0 else -1
        end = self.text.index(INSERT)
        row, col = index(end)
        # If the col is 0, set the index to the end of the previous row
        # (unless first row)
        if direction == -1:
            if row == 1 and col == 0:
                return "1.0", "1.0"
            elif row > 1 and col == 0:
                row, col = index(self.text.index("%d.end" % (row + direction)))
                end = index(row, col)
        # If the left char is whitespace, delete that AND the next word
        start = None
        if direction == -1:
            not_at_end = lambda x: x > 0
        else:
            _, end_point = index(self.text.index("%d.end" % row))
            not_at_end = lambda x: x < end_point
        while not_at_end(col):
            start = index(row, col + direction)
            char = self.text.get(start)
            # if the character is a space, delete that
            if char in " \t":
                start = index(row, col)
                break
            elif char in ".,()[]{}=\"'":
                start = index(row, col + (1 if direction == -1 else 0))
                break
            col = col + direction
        # return the "larger index" second
        if index(start) > index(end):
            start, end = end, start
        return start, end

    def delete_word(self, event):
        """ Deletes the preceeding text """
        if not self.delete_selection():
            start, end = self.look(-1)
            self.text.delete(start, end)
        self.update(event)
        # execute.update_line_numbers(self.text)
        return

    def delete_next_word(self, event):
        """ Deletes the following word """
        if not self.delete_selection():
            start, end = self.look(1)
            self.text.delete(start, end)
        self.update(event)
        return

    def delete_selection(self):
        """ If an area is selected, it is deleted and returns True """
        try:
            text = self.text.get(SEL_FIRST, SEL_LAST)
            a, b = self.text.index(SEL_FIRST), self.text.index(SEL_LAST)
            self.text.delete(SEL_FIRST, SEL_LAST)
            return True
        except Exception:
            return False

    def text_selected(self):
        """ Returns True if text is selected """
        try:
            self.text.index(SEL_FIRST)
            return True
        except Exception:
            return False

    """
        Keyboard Shortcuts
        ==================
    """
    # Select all: Ctrl+a
    # -------------------
    def select_all(self, event=None):
        """ Select the contents of the editor """
        self.text.tag_add(SEL, "1.0", END)
        self.text.mark_set(INSERT, "1.0")
        self.text.see(INSERT)
        return 'break'

    # Kill all: Ctrl+.
    # -----------------
    def killall(self, event=None):
        """ Stops all player objects """
        execute("_Clock.clear()", verbose=False)
        print("Clock.clear()")
        return "break"

    # Zoom in: Ctrl+=
    # ----------------
    def zoom_in(self, event=None):
        """ Ctrl+= increases text size """
        self.root.grid_propagate(False)
        font = tkFont.nametofont(self.codefont)
        size = font.actual()["size"]+2
        font.configure(size=size)
        # Increase size of line number
        self.linenumbers.config(width=self.linenumbers.winfo_width() + 3)
        return 'break'

    # Zoom out: Ctrl+-
    # -----------------
    def zoom_out(self, event=None):
        """ Ctrl+- decreases text size (minimum of 8) """
        self.root.grid_propagate(False)
        font = tkFont.nametofont(self.codefont)
        size = font.actual()["size"]-2
        if size >= 8:
            font.configure(size=size)
            self.linenumbers.config(width=self.linenumbers.winfo_width() - 3)
        return 'break'

    def submit(self, code_str):
        """ Runs the chunk of code through Renardo processing and execute """
        try:
            execute(code_str)
        except Exception as e:
            print(e)

    def highlight(self, start, end, colour="Red"):
        """ Highlights an area of text """
        # Label block (start and end are lines before & after the code itself)
        self.text.tag_add("code", start, end)
        # Highlight
        self.text.tag_config("code", background=colour, foreground="White")
        return

    def unhighlight(self):
        """
        Creates thread to wait 0.2 seconds before removing any highlights
        from the text
        """
        # Remove labels
        self.text.tag_delete("code")
        return

    def duplicate_line(self, event=None):
        """
        Called using Ctrl+D - duplicates the content of this line and moves
        the insert to the same place on the line below
        """
        # 1. get contents of line
        row, col = self.text.row_col(INSERT)
        line_start = "{}.0".format(row)
        line_end = "{}.{}".format(row, END)
        line = self.text.get(line_start, line_end)
        # 2. Add new line to end of this row
        self.text.insert(line_end, "\n{}".format(line))
        self.text.mark_set(INSERT, "{}.{}".format(row + 1, col))
        self.update_all()
        return "break"

    """
        Methods that view the Renardo namespace
        --------------------------------------
    """
    def update_prompt(self, visible=True):
        if visible:
            self.prompt.show()
        else:
            self.prompt.hide()
        return

    def update_prompt2(self):
        if self.inbrackets:
            # Show prompt
            line, column = index(self.text.index(INSERT))
            text = self.text.get(index(line, 0), index(line, column))
            x = self.font.measure(text)
            y = self.font.metrics("linespace") * line
            self.prompt.move(x, y)
            # If cursor is in between brackets that follow a type word
            try:
                self.check_namespace()
            except Exception:
                pass
        else:
            # Hide prompt
            self.prompt.hide()

    def check_namespace(self):
        """ Sets the label """
        obj = self.namespace[self.last_word]
        if obj:
            if obj.__doc__ is not None:
                self.prompt.value.set(obj.__doc__)
            else:
                self.prompt.hide()

    """
        Methods that update the contents of the IDE
        -------------------------------------------
    """
    def on_text_modified(self, event):
        self.text.modifying = not self.text.modifying
        if self.text.modifying:
            # Get number of lines added / removed
            old_length, new_length = self.text.lines, self.text.get_num_lines()
            dif = new_length - old_length
            # Get end index of the operation
            row, col = index(self.text.index(INSERT))
            if dif >= 0:
                for x in range(row - dif, row + 1):
                    self.colour_line(x)
            self.text.edit_modified(False)
        return "break"

    def update(self, event=None, insert=INSERT):
        """ Updates the the colours of the IDE """
        # Move the window to view the current line
        self.text.see(INSERT)
        self.update_prompt()
        return "break"

    def update_all(self):
        """ Updates every line in the IDE """
        row, col = index(self.text.index(END))
        lines = row + 1
        for line in range(lines):
            self.colour_line(line)
        return

    def colour_line(self, line):
        """
        Checks a line for any tags that match regex and updates IDE colours
        """
        start_of_line, end_of_line = index(line, 0), index(line, "end")
        thisline = self.text.get(start_of_line, end_of_line)
        try:
            # Remove tags at current point
            for tag_name in self.text.tag_names():
                if tag_name != "tag_open_brackets":
                    self.text.tag_remove(tag_name, start_of_line, end_of_line)
            # Re-apply tags
            for tag_name, start, end in findstyles(thisline):
                self.text.tag_add(tag_name,
                                  index(line, start),
                                  index(line, end))
        except Exception as e:
            print(e)
        # Find comments (not done with regex)
        i = find_comment(thisline)
        if i is not None:
            self.text.tag_add("comments", index(line, i), end_of_line)
        return

    def find_multiline(self):
        """
        Goes through the whole text and adds multiline formatting where
        necessary. Not-implemented
        """
        start = 1
        end = int(self.text.index(END).split(".")[0])
        pos = []
        tracking = False
        tracking_char = ""
        # Iterate over each line
        for line in range(start, end):
            # If it contains a """ or ''' start tracking
            text = self.text.get("{}.0".format(line), "{}.end".format(line))
            for char in ("'''", '"""'):
                print(char, text, char in text)
                if char in text:
                    index = text.index(char)
                    if tracking and tracking_char == char:
                        break
            else:
                index = None
            # Find the closing set of quotes
            if index is not None:
                if tracking:
                    tracking_char = ""
                    tracking = False
                    m_end = index
                    pos.append((m_start, m_end))
                else:
                    tracking = True
                    m_start = index
                    tracking_char = char
        # Add formatting
        if len(pos) > 0:
            print(pos)
        return

    def get_last_word(self):
        line, column = index(self.text.index(INSERT))
        string = ""
        while True:
            char = self.text.get(index(line, column-1))
            if char.isalpha():
                string = char + string
            else:
                break
            column -= 1
            if column == 0:
                break
        self.last_word = string
        return self.last_word

    def key_up(self, event=None):
        if self.prompt.visible:
            self.prompt.cycle_up()
            return "break"
        return

    def key_down(self, event=None):
        if self.prompt.visible:
            self.prompt.cycle_down()
            return "break"
        return

    def move_word_right(self, event=None, keep_selection=False):
        if not keep_selection:
            self.text.remove_selection()
        row, col = index(self.text.index(INSERT))
        searching = True
        if col == 0:
            _,  line_length = index(self.text.index("{}.end".format(row)))
            if line_length == 0:
                new_row, new_col = row + 1, 0
                searching = False
        col += 1
        last_index = self.text.index(END)
        chars = "()[]{} \"'.,"
        while searching:
            _, line_length = index(self.text.index("{}.end".format(row)))
            if line_length > 0:
                for i in range(col, line_length - 1):
                    if self.text.get(index(row, i)) in chars:
                        searching = False
                    elif index(row, i) == last_index:
                        searching = False
                    if searching is False:
                        new_row, new_col = row, i
                        self.text.mark_set(INSERT, index(new_row, new_col))
                        return "break"
                row += 1
                col = 0
            else:
                searching = False
                new_row, new_col = row, 0
        self.text.mark_set(INSERT, index(new_row, new_col))
        return "break"

    def move_word_left(self, event=None, keep_selection=False):
        if not keep_selection:
            self.text.remove_selection()
        row, col = index(self.text.index(INSERT))
        col -= 1
        last_index = self.text.index(END)
        chars = "()[]{} \"'.,"
        searching = True
        _, line_length = index(self.text.index("{}.end".format(row)))
        if line_length == 0:
            searching = False
            new_row, new_col = index(self.text.index("{}.end".format(row - 1)))
        while searching:
            if line_length > 0:
                for i in range(col, 0, -1):
                    if self.text.get(index(row, i)) in chars:
                        searching = False
                    elif index(row, i) == last_index:
                        searching = False
                    if searching is False:
                        new_row, new_col = row, i
                        self.text.mark_set(INSERT, index(new_row, new_col))
                        return "break"
                _, line_length = row, col = index(self.text.index("{}.end".format(row - 1)))
            else:
                searching = False
                new_row, new_col = row, 0
        self.text.mark_set(INSERT, index(new_row, new_col))
        return "break"

    def select_word_right(self, event=None):
        """ Calls self.move_word_right() and also selects the text moved """
        old, _, new = self.text.index(INSERT), self.move_word_right(keep_selection=True), self.text.index(INSERT)
        self.invert_selection(old, new)
        return "break"

    def select_word_left(self, event=None):
        old, _, new = self.text.index(INSERT), self.move_word_left(keep_selection=True), self.text.index(INSERT)
        self.invert_selection(old, new)
        return "break"

    def invert_selection(self, index1, index2=None):
        """
        Given two Tkinter indices, will select non-selected text in the range
        and de-select the selected text
        """
        for index in self.text.char_range(index1, index2):
            if self.text.is_selected(index):
                self.text.tag_remove(SEL, index)
            else:
                self.text.tag_add(SEL, index,)
        return

    """
        Generic functions
        -----------------
        - Correct exiting
        - Tabspace (todo: customise)
    """
    def kill(self):
        """ Proper exit function """
        self.terminate()
        self.root.destroy()
        return

    def tabspace(self):
        return " " * tabsize

    def terminate(self):
        """ Called on window close. Ends Clock thread process """
        execute("_Clock.stop()")
        execute("Server.quit()")
        return

    def releaseNodes(self, event=None):
        execute("Server.freeAllNodes()")
        return

    def replace(self, line, old, new):
        """ Replaces text on a specified line and updates the IDE """
        try:
            # Store cursor
            origin = self.text.index(INSERT)
            # Get contents of the line
            a = index(line, 0)
            b = index(line, END)
            contents = self.text.get(a, b)
            # Find the part to replace
            i = contents.index(old)
            j = i + len(old)
            a = index(line, i)
            b = index(line, j)
            self.text.delete(a, b)
            self.text.insert(a, new)
            self.update(insert=a)
            self.text.mark_set(INSERT, origin)
        except Exception:
            return

    def replace_re(self, line, new=""):
        """ Replaces text on a specified line and updates the IDE """
        try:
            # Store cursor
            origin = self.text.index(INSERT)
            # Get contents of the line
            a = index(line, 0)
            b = index(line, END)
            contents = self.text.get(a, b)
            # Search using RegEx
            match = re_pat(contents)
            if match:
                # Find the part to replace
                i = match.start()
                j = match.end()
                a = index(line, i)
                b = index(line, j)
                self.text.delete(a, b)
                self.text.insert(a, new)
                self.update(insert=a)
                self.text.mark_set(INSERT, origin)
        except Exception as e:
            print(e)
            return

    def set_all(self, text):
        self.text.delete("1.0", END)
        self.text.insert("1.0", text.strip())
        self.update_all()
        self.text.mark_set(INSERT, "1.0")
        return

    def get_all(self):
        """ Returns all the text as a string """
        return self.text.get("1.0", END).strip()

    def openhomepage(self):
        webbrowser.open("https://renardo.org/")
        return

    def opendocumentation(self):
        webbrowser.open("https://renardo.org/about/renardo/")
        return

    def set_temp_file(self, text):
        write_to_file(FOXDOT_TEMP_FILE, text)
        return

    def clear_temp_file(self):
        return

    def clear_console(self):
        self.console.clear()
        return

    def start_listening(self, **kwargs):
        """ Manual starting of Renardo tempo server """
        # ToDo - take this out of the menu
        self.listening_for_connections.set(not self.listening_for_connections.get())
        self.allow_connections(**kwargs)
        return

    def allow_connections(self, **kwargs):
        """
        Starts a new instance of ServerManager.TempoServer and connects it
        with the clock
        """
        Clock = self.namespace["_Clock"]
        if self.listening_for_connections.get() is True:
            Clock.start_tempo_server(TempoServer, **kwargs)
            print("Listening for connections on {}".format(Clock.tempo_server))
        else:
            Clock.kill_tempo_server()
            print("Closed connections")
        return

    def show_popup(self, *args):
        """ Shows the context menu when pressing right click """
        # Show text popup
        self.popup.show(*args)
        # Hide console popup
        self.console.popup.hide(*args)
        return

    def mouse_press(self, *args):
        """ De-select etc when pressing mouse 1 """
        self.console.canvas.select_clear()  # Clear select on the console
        self.popup.hide(*args)
        self.console.popup.hide(*args)
        return



from FoxDotEditor.tkimport import Text, SEL, END, SEL_FIRST, SEL_LAST, INSERT
from FoxDotEditor.Format import *

try:
    import Queue
except ImportError:
    import queue as Queue

background = colour_map['background']


class ThreadedText(Text):
    def __init__(self, master, **options):
        Text.__init__(self, master, **options)
        self.root = master
        self.config(
            highlightbackground=background,
            selectbackground="Dodger Blue",
            selectforeground="White",
            highlightthickness=0
        )
        self.height = options.get("height", 20)
        self.queue = Queue.Queue()
        self.lines = 0  # number of lines in the text
        self.modifying = False
        self.update()

    def on_resize(self, event):
        line_h = self.dlineinfo("@0,0")
        if line_h is not None:
            self.height = int((self.winfo_height()-2) / line_h[3])
        return

    def get_num_lines(self):
        self.lines = len(self.get("1.0", END).split("\n"))
        return self.lines

    def update(self):
        """ Recursively called method that monitors as
            queue of Tkinter tasks.
        """
        try:
            while True:
                task, args, kwargs = self.queue.get_nowait()
                task(*args, **kwargs)
                self.update_idletasks()
        # Break when the queue is empty
        except Queue.Empty:
            pass
        except Exception as e:
            print(e)
        # Recursive call
        self.after(10, self.update)
        return

    def has_selection(self):
        """ Returns True if the selection tag is present in the text box """
        return bool(self.tag_ranges(SEL))

    def remove_selection(self):
        """ Removes selection from the entire document """
        self.tag_remove(SEL, "1.0", END)
        return

    def is_selected(self, index):
        """ Returns True if the character at index has the SEL tag """
        return self.index(index) in self.char_range(SEL_FIRST, SEL_LAST) if self.has_selection() else False

    def row_col(self, index):
        return tuple([int(x) for x in self.index(index).split(".")])

    def is_after(self, index1, index2):
        """
        Returns True if index1 is after index2, returns True if they are equal
        """
        a_row, a_col = self.row_col(index1)
        b_row, b_col = self.row_col(index2)
        return (a_row > b_row) or (a_row == b_row and a_col >= b_col)

    def is_before(self, index1, index2):
        """
        Returns True if index1 is after index2, returns True if they are equal
        """
        return not self.is_after(index1, index2)

    def char_range(self, index1, index2):
        """ Returns a list of indices between two Tk indices"""
        if self.is_after(index1, index2):
            index1, index2 = index2, index1
            reverse = True
        else:
            reverse = False
        a_row, a_col = self.row_col(index1)
        b_row, b_col = self.row_col(index2)
        data = []
        for row in range(a_row, b_row + 1,):
            if row == a_row:
                x1_col = a_col
            else:
                x1_col = 0
            if row == b_row:
                x2_col = b_col
            else:
                _, x2_col = self.row_col(self.index("{}.{}".format(row, END)))

            for col in range(x1_col, x2_col):
                data.append("{}.{}".format(row, col))
        if reverse:
            data = list(reversed(data))
        return data

    def get_visible_range(self):
        """
        Returns a tuple of integers for the first and last row visible in the
        editor
        """
        a = self.index("@0,0")
        b = self.index("@0,%d" % self.winfo_height())
        return tuple(int(s.split(".")[0]) for s in (a, b))



from FoxDotEditor.tkimport import *
from FoxDotEditor.tkimport import Text, SEL, END, SEL_FIRST, SEL_LAST, INSERT
from FoxDotEditor.Format import *


class SearchBar:

    def __init__(self, parent):
        self.root = parent.root
        self.parent = parent
        self.f_height = 50
        self.sb_frame = tb.Frame(
            self.root, height=self.f_height)
        self.sb_frame.grid(row=1, column=2, sticky='ew')
        self.hide()
        self.search_list = list()
        self.search = ""
        self.idx = ""
        self.sb_label = tb.Label(
            self.sb_frame, text="Enter search")
        self.sb_label.grid(row=0, column=0, padx=10)
        self.search_entry = tb.Entry(
            self.sb_frame, width=40, justify="left")
        self.search_entry.grid(row=0, column=1, padx=10)
        self.search_btn = tb.Button(
            self.sb_frame, text="Search", command=self.search_task)
        self.search_btn.grid(row=0, column=2, padx=10, pady=5, sticky="e")

    def reset_list(self):
        if self.search != self.search_entry.get():
            self.search_list.clear()
            self.parent.text.tag_remove(SEL, "1.0", "end-1c")

    def search_task(self):
        self.reset_list()
        self.parent.text.focus_set()
        self.search = self.search_entry.get()

        if self.search:
            if self.search_list == []:
                self.idx = "1.0"
            else:
                self.idx = self.search_list[-1]
            self.idx = self.parent.text.search(self.search, self.idx, nocase=1,
                                               stopindex=END)
            self.lastidx = '%s+%dc' % (self.idx, len(self.search))
            try:
                self.parent.text.tag_remove(SEL, "1.0", self.lastidx)
            except Exception:
                pass
            try:
                self.parent.text.tag_add(SEL, self.idx, self.lastidx)
                self.counter_list = []
                self.counter_list = str(self.idx).split('.')
                self.parent.text.mark_set("insert", "%d.%d" % (float(int(self.counter_list[0])), float(int(self.counter_list[1]))))
                self.parent.text.see(float(int(self.counter_list[0])))
                self.search_list.append(self.lastidx)
            except Exception:
                tkMessageBox.showinfo("Search complete", "No further matches")
                self.search_list.clear()
                self.parent.text.tag_remove(SEL, "1.0", "end-1c")

    def hide(self):
        """ Removes searchbar from interface """
        self.sb_frame.grid_remove()
        return

    def show(self):
        self.sb_frame.grid()
        return



from FoxDotEditor.tkimport import *
from renardo_lib.Settings import LINE_NUMBER_MARKER_OFFSET
from renardo_lib.Code import execute


class LineNumbers(tb.Canvas):
    def __init__(self, master, *args, **kwargs):
        tb.Canvas.__init__(self, *args, **kwargs)
        self.root = master
        self.textwidget = master.text

    def redraw(self, *args):
        '''redraw line numbers'''
        # Update player line numbers
        # execute.update_line_numbers(self.textwidget)
        # Clear
        self.delete("all")
        # Draw a line
        w = self.winfo_width() - 1
        h = self.winfo_height()
        self.create_line(w, 0, w, h, fill="gray")
        i = self.textwidget.index("@0,0")

        while True:
            dline = self.textwidget.dlineinfo(i)
            if dline is None:
                break
            y = dline[1]
            h = dline[3]
            linenum = int(str(i).split(".")[0])
            curr_row = int(self.textwidget.index(INSERT).split(".")[0])

            if linenum == curr_row:
                x1, y1 = 0, y + LINE_NUMBER_MARKER_OFFSET
                x2, y2 = w - 2, y + h
                self.create_rectangle(x1,
                                      y1,
                                      x2,
                                      y2,
                                      fill="gray30",
                                      outline="gray30")

            self.create_text(w - 4, y, anchor="ne",
                             justify=RIGHT,
                             text=linenum,
                             font=self.root.codefont,
                             fill="#c9c9c9")

            i = self.textwidget.index("{}+1line".format(i))

        # Update console beat counter here too
        self.root.console.counter.redraw()
        self.after(30, self.redraw)

    def hide(self):
        """ Removes treeview from interface """
        self.grid_remove()
        return

    def show(self):
        self.grid()
        return



#!/usr/bin/env python
from __future__ import absolute_import

import multiprocessing
import os
from .tkimport import *
from .Format import *
from renardo_lib.Settings import *
from renardo_gatherer.collections import nonalpha, SAMPLES_DIR_PATH

try:
    from playsound import playsound
except Exception:
    print("playsound library not installed...")


class SampleChart:

    def __init__(self):
        # Basic TKinter function
        self.root = Tk()
        self.width = 800
        self.height = 600
        self.wheel_count = 0
        self.root.minsize(self.width, self.height)
        self.root.title("FoxDot >> Samples Database Chart")
        self.root.resizable(True, True)
        self.root.grid_rowconfigure(0, weight=1)  # configure grid system
        self.root.grid_columnconfigure(0, weight=1)
        # self.root.iconbitmap('img/foxdot.ico')
        # Call init methods
        self.sp_count = len([name for name in os.listdir(str(SAMPLES_DIR_PATH))])
        self.sp_names = os.listdir(str(SAMPLES_DIR_PATH))
        self.sp_names.sort()
        self.sp_view()
        self.create_dics(spack_id="0_foxdot_default")
        self.smpl_view(self.width, self.height)
        self.root.mainloop()

    def sp_view(self):
        """Generate a button for each sample database by spack folder"""
        # Set SoundPack_SAMPLES Frame
        self.sp_frame = Frame(self.root)
        self.sp_frame.grid(row=0, column=0, padx=10, pady=10, sticky="nswe")
        # Label for SoundPack List
        self.sp_label = Label(
            self.sp_frame,
            text="SAMPLES_PACK",
            compound="left")
        self.sp_label.grid(row=0, column=0, padx=10, sticky="w")
        # Add Buttons
        for count in range(self.sp_count):
            lbl = self.sp_names[count]
            btn_sp = Button(
                self.sp_frame,
                text=lbl,
                command=lambda sp=lbl: self.change_sp(sp))
            btn_sp.grid(row=1, column=count, padx=10, pady=10, sticky="w")
        self.txt_frame = Frame(self.sp_frame)
        self.txt_frame.grid(columnspan=self.sp_count,
                            row=2, column=0, padx=20, pady=10, sticky="nsew")
        self.lbl_file = Label(
            self.txt_frame,
            text="Filename: ",
            anchor="w")
        self.lbl_file.grid(columnspan=self.sp_count,
                           row=0,
                           column=0,
                           padx=20,
                           pady=5,
                           sticky="w")
        self.txt = Text(self.txt_frame, width=80, height=1)
        self.txt.grid(row=1, column=0, padx=20, pady=20, sticky="nsew")

    def create_dics(self, spack_id="0_foxdot_default"):
        """
        Iterating through subfolders and writing all file names into dictionary
        lists
        """
        self.spack_id_str = str(spack_id)
        self.ext = ('.wav', 'aif')
        self.dict_letters = {}
        self.dict_letters.clear()
        self.dict_specials = {}
        self.dict_specials.clear()
        self.dict_loops = []
        # First go through all letters and get file paths in upper and lower
        # Fill dictionary with letters as key and file names of audio as values
        self.sp_path_l = str(SAMPLES_DIR_PATH) + "/" + spack_id + "/"
        self.dir_list_l = []

        for filename in os.listdir(self.sp_path_l):
            if os.path.isdir(os.path.join(self.sp_path_l, filename)):
                self.dir_list_l.append(filename)

        self.dir_list_l.sort()

        for i in self.dir_list_l:
            if i != "_" and i != "_loop_":
                self.new_path = self.sp_path_l + str(i) + "/lower/"
                # self.smpl_list = os.path.isdir(self.new_path)
                self.smpl_list = [
                    f for f in os.listdir(self.new_path)
                    if os.path.isfile(os.path.join(self.new_path, f))
                ]
                self.smpl_list.sort()
                for n in self.smpl_list:
                    if not n.endswith(self.ext):
                        self.smpl_list.remove(n)
                self.dict_letters[i.upper()] = self.smpl_list
                self.dict_letters[i] = self.smpl_list
                self.new_path = self.sp_path_l + str(i) + "/upper/"
                self.smpl_list = [
                    f for f in os.listdir(self.new_path)
                    if os.path.isfile(os.path.join(self.new_path, f))
                ]
                self.smpl_list.sort()
                for n in self.smpl_list:
                    if not n.endswith(self.ext):
                        self.smpl_list.remove(n)
                self.dict_letters[i.upper()] = self.smpl_list

        # Fill dictionary with specials as key and file names of
        # audio as values
        self.sp_path_s = self.sp_path_s = str(SAMPLES_DIR_PATH)+"/"+spack_id+"/_/"
        self.dir_list_s = []
        for filename in os.listdir(self.sp_path_s):
            if os.path.isdir(os.path.join(self.sp_path_s, filename)):
                self.dir_list_s.append(filename)
        self.dir_list_s.sort()
        for j in self.dir_list_s:
            self.new_path = self.sp_path_s + str(j) + "/"
            self.smpl_list = [
                f for f in os.listdir(self.new_path)
                if os.path.isfile(os.path.join(self.new_path, f))
            ]
            self.smpl_list.sort()
            for n in self.smpl_list:
                if not n.endswith(self.ext):
                    self.smpl_list.remove(n)
            self.dict_specials[j] = self.smpl_list
        # Fill dictionary with loops as value
        self.sp_path_loops = str(SAMPLES_DIR_PATH)+"/"+spack_id+"/"+FOXDOT_LOOP
        self.smpl_list = [
            f for f in os.listdir(self.sp_path_loops)
            if os.path.isfile(os.path.join(self.sp_path_loops, f))
        ]
        self.smpl_list.sort()
        for n in self.smpl_list:
            if not n.endswith(self.ext):
                self.smpl_list.remove(n)
        self.dict_loops = self.smpl_list

    def smpl_view(self, w, h):
        """Generate a button for each sample arranged by character folder"""
        self.w = self.width
        self.h = self.height
        self.processes = []
        # Create basic frame that will contain all
        self.frame = Frame(self.root)
        self.frame.grid(row=2, column=0, sticky='news')
        self.frame.grid_columnconfigure(0, weight=1)
        self.frame_label = Label(self.frame,
                                 text="SAMPLES",
                                 compound="left")
        self.frame_label.grid(row=0, column=0, padx=20, sticky="w")
        # Add a canvas in that frame
        self.canvas = Canvas(self.frame, bg=colour_map['background'])
        self.canvas.grid(row=1, column=0, sticky='news')
        # Link a scrollbar to the canvas
        self.xsb = Scrollbar(self.frame, orient="horizontal",
                             command=self.canvas.xview)
        self.ysb = Scrollbar(self.frame, orient="vertical",
                             command=self.canvas.yview)
        self.xsb.grid(row=2, column=0, sticky='we')
        self.ysb.grid(row=1, column=1, sticky='ns')
        self.canvas.configure(xscrollcommand=self.xsb.set,
                              yscrollcommand=self.ysb.set)
        self.canvas.grid_rowconfigure(0, weight=1)
        self.canvas.grid_rowconfigure(1, weight=1)
        self.canvas.grid_columnconfigure(0, weight=1)
        self.canvas.grid_columnconfigure(1, weight=1)
        # Create a frame to contain Buttons
        self.btn_frame = Frame(self.canvas, bg=colour_map['background'])
        self.canvas.create_window((0, 0), window=self.btn_frame, anchor='nw')
        # Generate sample buttons
        self.smpl_btns()
        # Update buttons frames idle tasks to let tkinter calculate
        # buttons sizes
        self.btn_frame.update_idletasks()
        self.canvas['xscrollcommand'] = self.xsb.set
        # Resize the canvas frame
        self.canvas.config(width=self.w - self.ysb.winfo_width(),
                           height=self.h - self.xsb.winfo_height())
        # Set the canvas scrolling region
        self.canvas.config(scrollregion=self.canvas.bbox("all"))
        self.canvas.grid_propagate(True)
        # This is what enables using the mouse:
        self.btn_frame.bind("<ButtonPress-1>", self.move_start)
        self.btn_frame.bind("<B1-Motion>", self.move_move)
        # linux scroll
        self.btn_frame.bind("<Button-4>", self.on_mousewheel)
        self.btn_frame.bind("<Button-5>", self.on_mousewheel)
        self.btn_frame.bind("<Prior>", self.on_mousewheel)  # Bind to PageUp
        self.btn_frame.bind("<Next>", self.on_mousewheel)  # Bind to PageDown
        # play audio
        # self.root.bind("<space>", self.space_play)
        # windows scroll
        self.btn_frame.bind("<MouseWheel>", self.on_mousewheel)
        self.btn_frame.bind("<space>", self.press_space)
        # Update canvas when changing window size
        self.root.bind("<Configure>", self.on_resize)

    def smpl_btns(self):
        """Generate audio sample buttons"""
        self.colors = list(colour_map.keys())
        self.col_space = 12
        # First delete all in btn_frame
        for widgets in self.btn_frame.winfo_children():
            widgets.destroy()
        # Add category buttons
        self.counter = 0
        for k in self.dict_letters.keys():
            self.btn = Button(self.btn_frame, text=k, width=10,
                              fg='white', bg=colour_map['background'])
            self.btn.grid(row=self.counter, column=0)
            self.counter += 1
        for n in self.dict_specials.keys():
            self.special = list(nonalpha.keys())[list(nonalpha.values()).index(n)]
            self.btn = Button(self.btn_frame, text=self.special,
                              width=10, fg='white',
                              bg=colour_map['background'])
            self.btn.grid(row=self.counter, column=0)
            self.counter += 1
        self.btn = Button(self.btn_frame, text="loop", width=10,
                          fg='white', bg=colour_map['background'])
        self.btn.grid(row=self.counter, column=0)
        # Create buttons for each sample in the letters dictionary
        self.dcounter = 0
        for m in self.dict_letters:
            self.fcounter = 0
            for v in self.dict_letters[m]:
                self.word = v.lower()
                if "_" in self.word or "-" in self.word:
                    self.word = self.word.replace('.', '_')
                    self.word = self.word.replace('-', '_')
                    self.word = self.word.split('_')
                self.color = ""
                self.btn = Button(self.btn_frame,
                                  text=self.fcounter,
                                  width=1,
                                  command=lambda r=m,
                                  c=self.fcounter,
                                  p=v,
                                  d=self.dcounter,
                                  f=self.fcounter+self.col_space: self.play_audio(r, c, p, d, f))
                for c in self.colors:
                    if c in self.word:
                        self.color = c
                        self.btn.configure(bg=colour_map[self.color])
                    elif self.color == "":
                        self.btn.configure(bg=colour_map["default"])
                self.btn.grid(row=self.dcounter, column=self.fcounter + self.col_space)
                self.fcounter += 1
            self.dcounter += 1
        # Create buttons for each sample in the specials dictionary
        for n in self.dict_specials:
            self.fcounter = 0
            for w in self.dict_specials[n]:
                self.word = w.lower()
                if "_" in self.word or "-" in self.word:
                    self.word = self.word.replace('.', '_')
                    self.word = self.word.replace('-', '_')
                    self.word = self.word.split('_')
                self.color = ""
                self.btn = Button(self.btn_frame,
                                  text=self.fcounter,
                                  width=1,
                                  command=lambda r=n,
                                  c=self.fcounter,
                                  p=w,
                                  d=self.dcounter,
                                  f=self.fcounter+self.col_space: self.play_audio(r, c, p, d, f))
                for c in self.colors:
                    if c in self.word:
                        self.color = c
                        self.btn.configure(bg=colour_map[self.color])
                    elif self.color == "":
                        self.btn.configure(bg=colour_map["default"])
                self.btn.grid(row=self.dcounter, column=self.fcounter + self.col_space)
                self.fcounter += 1
            self.dcounter += 1
        self.fcounter = 0
        for path in self.dict_loops:
            self.btn = Button(self.btn_frame,
                              text=self.fcounter,
                              width=1,
                              fg="white",
                              bg=colour_map["loops"],
                              command=lambda r="loops",
                              c=self.fcounter,
                              p=path,
                              d=self.dcounter,
                              f=self.fcounter+self.col_space: self.play_audio(r, c, p, d, f))
            # self.btn_frame.bind('<Return>', lambda event=None: button.invoke())
            self.btn.grid(row=self.dcounter, column=self.fcounter + self.col_space)
            self.fcounter += 1

    def play_audio(self, char, sample, path, row, col):
        """Displays sample code to copy and plays audio"""
        if len(self.processes) > 0:
            for p in self.processes:
                p.terminate()
        self.info = [row, col]
        self.txt.delete('1.0', END)
        self.char = char
        self.cmd = ""

        if len(self.char) == 1 and self.char.isalpha():
            if self.char.isupper():
                self.path = self.sp_path_l + self.char.lower() + "/upper/" + path
            elif self.char.islower():
                self.path = self.sp_path_l + self.char + "/lower/" + path
            self.cmd = "play"
        elif self.char == "loops":
            self.path = self.sp_path_l + FOXDOT_LOOP + "/" + path
            self.cmd = "loop"
            self.char = os.path.splitext(path)[0]
        else:
            self.char = list(nonalpha.keys())[list(
                nonalpha.values()).index(self.char)]
            self.path = self.sp_path_s + char + "/" + path
            self.cmd = "play"
        try:
            self.p = multiprocessing.Process(target=playsound, args=(self.path,))
        except Exception:
            print("playsound library not installed...")
        if self.spack_id_str == "0_foxdot_default":
            spack = "0"
        else:
            spack = self.spack_id_str
        self.code = f'{self.cmd}("{self.char}", spack={spack}, sample={sample})'
        self.txt.insert(END, self.code)
        self.change_fname(path)
        self.p.start()
        self.processes.append(self.p)

    def press_space(self, event):
        # self.play_audio(char, sample, path, row, col)
        pass

    def change_sp(self, spack_id):
        """Deletes buttons and regenerate dictionaries and buttons with set database number"""
        self.sp_id = spack_id
        self.create_dics(self.sp_id)
        self.smpl_btns()
        # Update buttons frames idle tasks to let tkinter calculate buttons sizes
        self.btn_frame.update_idletasks()
        # Resize the canvas frame
        self.canvas.config(width=self.w - self.ysb.winfo_width(),
                           height=self.h - self.xsb.winfo_height())
        # Set the canvas scrolling region
        self.canvas.config(scrollregion=self.canvas.bbox("all"))
        self.canvas.grid_propagate(True)

    def change_fname(self, path):
        self.lbl_file["text"] = "Filename: " + path

    def on_mousewheel(self, event):
        """respond to Linux or Windows wheel event"""
        if event.num == 5 or event.delta == -120:
            self.wheel_count -= 1
        if event.num == 4 or event.delta == 120:
            self.wheel_count += 1

    def move_start(self, event):
        self.canvas.scan_mark(event.x, event.y)

    def move_move(self, event):
        self.canvas.scan_dragto(event.x, event.y, gain=1)

    def on_resize(self, event):
        self.width = self.root.winfo_width() - 10
        self.height = self.root.winfo_height() - self.sp_frame.winfo_reqheight() - 30
        # resize the canvas
        self.canvas.config(width=self.width - self.ysb.winfo_width(),
                           height=self.height - self.xsb.winfo_height())



# !/usr/bin/python
from FoxDotEditor.tkimport import *
from FoxDotEditor.Format import *
from FoxDotEditor.AppFunctions import stdout
from FoxDotEditor.MenuBar import ConsolePopupMenu
import math
import random
try:
    import Queue
except ImportError:
    import queue as Queue

""" Console widget that displays the true Python input """


class console:

    def __init__(self, parent):
        self.app = parent
        self.root = parent.root
        self.y_scroll = tb.Scrollbar(self.root)
        self.y_scroll.grid(row=3, column=3, sticky='nsew', rowspan=2)
        self.scrollable = False
        # Right-click menu
        self.popup = ConsolePopupMenu(self)
        # Create a bar for changing console size and displaying info about
        # beat number
        self.drag = tb.Frame(
            self.root,
            cursor="sb_v_double_arrow",
            height=3
        )
        self.counter = Counter(
            self,
            self.root,
            bd=0,
            bg='black',
            height=25,
            highlightthickness=0
        )
        # Create canvas
        self.height = 10
        self.max_offset = 0
        self.root_h = self.height + self.app.text.height
        self.canvas = tb.Canvas(
            self.root,
            bg=colour_map['console_bg'],
            bd=0,
            height=200,
            yscrollincrement=1,
            highlightthickness=0)

        self.canvas.bind("<Button-1>",
                         self.canvas_mouseclick)
        self.canvas.bind("<ButtonRelease-1>",
                         self.canvas_mouserelease)
        self.canvas.bind("<B1-Motion>",
                         self.canvas_mousedrag)
        self.canvas.bind("<MouseWheel>",
                         self.on_scroll)
        self.canvas.bind("<Button-{}>".format(2 if SYSTEM == MAC_OS else 3),
                         self.show_popup)
        self.canvas.bind("<{}-c>".format("Command" if SYSTEM == MAC_OS else "Control"),
                         self.edit_copy)
        self.padx = 5
        self.pady = 5
        self.text_y = 0
        self.text_height = 0
        self.canvas_height = 0
        # Draw logo
        # self.draw_logo()
        self.hello_msg = "Welcome to Renardo! Press Ctrl/Cmd + H for help."
        self.hello_lines = "-" * len(self.hello_msg)
        self.hello_txt = f"{self.hello_msg}\n{self.hello_lines}\n"
        # Create text
        self.text = self.canvas.create_text((self.padx, self.pady),
                                            anchor=NW,
                                            fill=colour_map['console_text'],
                                            font=self.app.console_font,
                                            text=self.hello_txt)
        self.text_cursor = None
        self.y_scroll.config(command=self.scroll_text)
        # Allow for resizing
        self.mouse_down = False
        self.drag.bind("<Button-1>",
                       self.drag_mouseclick)
        self.drag.bind("<ButtonRelease-1>",
                       self.drag_mouserelease)
        self.drag.bind("<B1-Motion>",
                       self.drag_mousedrag)
        self.drag.grid(row=2, column=1, stick="nsew", columnspan=3)
        self.canvas.grid(row=3, column=1, sticky="nsew", columnspan=2)
        self.counter.grid(row=4, column=1, sticky="nsew", columnspan=2)
        self.counter.hide()
        self.queue = Queue.Queue()
        self.update()

    def __str__(self):
        """ str(s) -> string """
        return self.canvas.itemcget(self.text, "text")

    def clear(self):
        """ Clears the console """
        self.canvas.dchars(self.text, 0, "end")
        return

    def configure(self, *args, **kwargs):
        self.canvas.config(*args, **kwargs)
        return

    def flush(self):
        return

    def write(self, string):
        """ Adds string to the bottom of the console """
        self.queue.put(string)
        return

    def drag_mouseclick(self, event):
        """ Allows the user to resize the console height """
        self.mouse_down = True
        self.root.grid_propagate(False)
        return

    def drag_mouserelease(self, event):
        self.mouse_down = False
        self.app.text.focus_set()
        return

    def drag_mousedrag(self, event):
        if self.mouse_down:
            textbox_line_h = self.app.text.dlineinfo("@0,0")
            if textbox_line_h is not None:
                self.app.text.height = int(self.app.text.winfo_height() /
                                           textbox_line_h[3])
            self.root_h = self.height + self.app.text.height
            widget_y = self.canvas.winfo_rooty()
            new_height = (self.canvas.winfo_height() +
                          (widget_y - event.y_root))
            self.height, old_height = new_height, self.height
            self.canvas.configure(height=max(self.height, 200))
            return "break"

    def update(self):
        try:
            while True:
                # Add last "print" to the console text
                string = self.queue.get_nowait()
                self.canvas.itemconfig(self.text,
                                       width=self.canvas.winfo_width())
                self.canvas.insert(self.text, "end", string)
                # Get the text bounding box
                bbox = self.canvas.bbox(self.text)
                # Text box height
                self.text_height = bbox[3] - bbox[1]
                # Canvas height
                self.canvas_height = self.canvas.winfo_height()
                # Only allow scrolling when the text is larger than the canvas
                if self.text_height > self.canvas_height:
                    self.scrollable = True
                    # The text should only move so that the end is at the
                    # bottom of the canvas
                    self.max_offset = self.canvas_height - self.text_height
                    self.text_y = self.max_offset
                    self.canvas.coords(self.text, (self.padx, self.text_y))
                else:
                    self.scrollable = False
                self.update_scrollbar()
        except Queue.Empty:
            pass
        self.root.after(50, self.update)

    def read(self):
        """ Returns contents of the console widget """
        return self.canvas.itemcget(self.text, "text")

    def hide(self):
        """ Removes console from interface """
        self.canvas.grid_remove()
        self.y_scroll.grid_remove()
        return

    def show(self):
        self.canvas.grid()
        self.y_scroll.grid()
        return

    def canvas_mouseclick(self, event):
        """ Forces the text to align itself and gives focus to the console """
        self.canvas.insert(self.text, "end", "")
        self.canvas.focus_set()
        self.canvas.focus(self.text)
        # Remove current selection
        self.canvas.select_clear()
        # Calculate current mouse pos
        x = self.canvas.canvasx(event.x)
        y = self.canvas.canvasx(event.y)
        self.text_cursor = "@%d,%d" % (x, y)
        return

    def canvas_mousedrag(self, event):
        """ Changes selection """
        x = self.canvas.canvasx(event.x)
        y = self.canvas.canvasx(event.y)
        xy = "@%d,%d" % (x, y)
        self.canvas.select_from(self.text, self.text_cursor)
        self.canvas.select_to(self.text, xy)
        return

    def canvas_mouserelease(self, event):
        self.text_cursor = None
        return

    def edit_copy(self, event=None):
        if self.canvas.select_item() == self.text:
            self.root.clipboard_clear()
            a = self.canvas.index(self.text, SEL_FIRST)
            b = self.canvas.index(self.text, SEL_LAST) + 1
            self.root.clipboard_append(self.canvas.itemcget(self.text,
                                                            "text")[a:b])
        return "break"

    def select_all(self, event=None):
        self.canvas.select_from(self.text, 0)
        self.canvas.select_to(self.text, "end")
        return "break"

    def scroll_text(self, *args):
        # Moves the text by an amount (pressing the arrow buttons)
        if args[0] == "scroll":
            self.move_text(int(args[1])*-1000)
        # Dragging the scroll bar
        elif args[0] == "moveto":
            new_y = float(args[1])
            if new_y > 1:
                new_y = 1
            elif new_y < 0:
                new_y = 0
            a, b = self.update_scrollbar(new_y)
            size = b - a
            try:
                self.text_y = (new_y / (1 - size)) * self.max_offset + self.pady
            except ZeroDivisionError:
                self.text_y = 0
            self.canvas.coords(self.text, (self.padx, self.text_y))
        return

    def move_text(self, delta):
        """ Moves the text up (negative) or down (positive) """
        if SYSTEM != MAC_OS:
            delta /= 100
        x, y = self.canvas.coords(self.text)
        self.text_y = max(min(self.pady, y + delta), self.max_offset)
        self.canvas.coords(self.text, (x, self.text_y))
        self.update_scrollbar()
        return

    def on_scroll(self, event):
        if self.scrollable:
            self.move_text(event.delta)
        return "break"

    def get_scrollbar_size(self):
        return float(self.canvas_height) / self.text_height

    def update_scrollbar(self, point=None):
        """ point should be a value between 0 and 1.0 """
        if not self.scrollable:
            a, b = 0.0, 1.0
        elif self.text_height != 0 and self.max_offset != 0:
            size = self.get_scrollbar_size()
            if point is not None:
                a = point
            else:
                a = (float(self.text_y) / self.max_offset) * (1-size)
            b = a + size
        self.y_scroll.set(a, b)
        return a, b

    def draw_arrow(self, start_x, start_y, width, colour, direction, degree=45):
        """
        Works out the line to draw at 45 degrees, returns the x and y of the
        end of the line. Direction should be a string, "up" or "down"
        """
        # Work out the height
        height = int(math.tan(math.radians(degree)) * (width / 2.0))
        # Draw line up
        end_x = start_x + (width / 2.0)
        end_y = (start_y - height) if direction == "up" else (start_y + height)
        self.canvas.create_line((start_x, start_y, end_x, end_y),
                                fill=colour,
                                width=3)
        # Draw line down
        start_x, start_y = end_x, end_y
        end_x = start_x + (width / 2.0)
        end_y = (start_y + height) if direction == "up" else (start_y - height)
        self.canvas.create_line((start_x, start_y, end_x, end_y),
                                fill=colour,
                                width=3)
        # Return the end x, y
        return end_x, end_y

    def draw_logo(self):
        """ Draws the red and green lines in the bg of the console.
         Future versions will have randomly created lines & possibly
         animated.
        """
        # All lines are 45 degress up then down
        grn_widths = [
            random.randint(400, 600),  # Large
            random.randint(150, 350),  # Medium
            random.randint(25, 100),   # Small
            ]
        # Shuffle the widths and use a mirrored version for red
        random.shuffle(grn_widths)
        red_widths = reversed(grn_widths)
        start_x = random.choice([50, 100, 150, 200])
        start_y = random.choice([50, 100])
        step = random.choice([10, 20, 25])
        # Draw Red line
        x, y = start_x, start_y
        for w in red_widths:
            x, y = self.draw_arrow(x, y, w, '#571d0c', "down")
            x += step
        # Draw Green line
        # Define start point
        x, y = start_x, start_y + 100
        for w in grn_widths:
            x, y = self.draw_arrow(x, y, w, '#1d5335', "up")
            x += step
        return

    def show_popup(self, *args):
        """ Shows the right-click context menu and hides the text popup """
        self.popup.show(*args)
        self.app.popup.hide(*args)


class Counter(Canvas):
    def __init__(self, parent, *args, **kwargs):
        Canvas.__init__(self, *args, **kwargs)
        self.parent = parent
        self.metro = self.parent.app.namespace['Clock']
        self.font = self.parent.app.console_font
        self.bg = colour_map.get('background', )
        self.active = True

    def hide(self):
        self.active = False
        self.grid_remove()

    def unhide(self):
        self.active = True
        self.grid()

    def toggle(self):
        if self.active:
            self.hide()
        else:
            self.unhide()

    def redraw(self):
        """
        Draw boxes and highlight current beat
        """
        if not self.active:
            return
        cycle = self.metro.meter[0]
        # Need to adjust for latency
        beat = int((self.metro.now() - self.metro.get_latency()) % cycle)
        w = self.winfo_width()
        h = self.winfo_height()
        self.delete("all")
        width = 120
        box_width = width / cycle
        h_offset = 8
        box_height = h - h_offset
        for n in range(cycle):
            x1, x2 = [(val*box_width)+(w-width-35) for val in [n, (n+1)]]
            y1, y2 = h_offset / 2, box_height + (h_offset / 2)
            bg = "red" if n == beat else self.bg
            self.create_rectangle(x1, y1, x2, y2)
            # self.create_rectangle(x1, y1, x2, y2, fill=bg, outline=self.bg, )
        self.create_text(x2 + (w - x2)/2,
                         h / 2,
                         justify=RIGHT,
                         text=beat + 1,
                         font=self.font,
                         fill="#c9c9c9")



"""
    Simple File System Explorer with Tk
"""
from FoxDotEditor.tkimport import *
from pathlib import Path
from renardo_lib.Settings import FOXDOT_EDITOR_ROOT
import os


class TreeView:

    def __init__(self, parent):
        self.app = parent
        self.root = parent.root
        self.tv_pane = tb.Frame(
            self.root,
            width=220,
            height=400)
        self.tv_pane.grid(row=0, column=0, sticky='nsew', rowspan=4)
        # Make sure the treeview widget follows the window
        # when resizing.
        self.tv_pane.rowconfigure(0, weight=1)
        self.tv_pane.columnconfigure(0, weight=1)
        self.hide()
        self.y_scroll = tb.Scrollbar(self.tv_pane)
        self.y_scroll.grid(row=0, column=1, sticky='ns')
        self.scrollable = False
        self.drag = tb.Frame(
            self.tv_pane,
            cursor="sb_h_double_arrow",
            width=5
        )
        # show="tree" removes the column header, since we
        # are not using the table feature.
        self.treeview = tb.Treeview(self.tv_pane, show="tree")
        self.treeview.grid(row=0, column=0, sticky="nsew", rowspan=3)
        # Call the item_opened() method each item an item
        # is expanded.
        self.treeview.tag_bind(
            "fstag", "<<TreeviewOpen>>", self.item_opened)
        # This dictionary maps the treeview items IDs with the
        # path of the file or folder.
        self.fsobjects: dict[str, Path] = {}
        self.file_image = PhotoImage(file=FOXDOT_EDITOR_ROOT+"/img/file.png")
        self.folder_image = PhotoImage(file=FOXDOT_EDITOR_ROOT+"/img/folder.png")
        # Load the user directory.
        self.home_path = str(Path.home())
        self.load_tree(Path(os.path.expanduser(self.home_path)))
        self.mouse_down = False
        self.drag.bind("<Button-1>",
                       self.drag_mouseclick)
        self.drag.bind("<ButtonRelease-1>",
                       self.drag_mouserelease)
        self.drag.bind("<B1-Motion>",
                       self.drag_mousedrag)
        self.drag.grid(row=0, column=2, stick="ns")
        self.treeview.bind("<Double-Button-1>", self.doubleclick)
        self.treeview.bind("<ButtonRelease-1>", self.drag_mouserelease)
        self.treeview.grid(row=0, column=0, stick="ns")

    def doubleclick(self, event):
        self.mouse_down = True
        item = self.treeview.selection()
        parent_iid = self.treeview.parent(item)
        node = []
        # go backward until reaching root
        while parent_iid != '':
            node.insert(0, self.treeview.item(parent_iid)['text'])
            parent_iid = self.treeview.parent(parent_iid)
        i = self.treeview.item(item, "text")
        path = self.home_path + os.path.sep + os.path.join(*node, i)
        if path.endswith(".py"):
            self.app.opentvfile(path)
        return

    def drag_mouseclick(self, event):
        """ Allows the user to resize the treeview width """
        self.mouse_down = True
        self.root.grid_propagate(False)
        return

    def drag_mouserelease(self, event):
        self.mouse_down = False
        return

    def drag_mousedrag(self, event):
        if self.mouse_down:
            self.root_v = self.tv_pane.winfo_width()
            widget_x = self.tv_pane.winfo_rootx()
            new_width = (self.tv_pane.winfo_width() + (widget_x - event.x_root))
            self.width, old_width = new_width, self.tv_pane.winfo_width()
            self.tv_pane.configure(width=max(self.width, 200))
            # return "break"

    def safe_iterdir(self, path: Path) -> tuple[Path, ...] | tuple[()]:
        """
        Like `Path.iterdir()`, but do not raise on permission errors.
        """
        try:
            return tuple(path.iterdir())
        except PermissionError:
            print("You don't have permission to read", path)
            return ()

    def get_icon(self, path: Path):
        """
        Return a folder icon if `path` is a directory and
        a file icon otherwise.
        """
        return self.folder_image if path.is_dir() else self.file_image

    def insert_item(self, name: str, path: Path, parent: str = "") -> str:
        """
        Insert a file or folder into the treeview and return the item ID.
        """
        iid = self.treeview.insert(
            parent, END, text=name, tags=("fstag",),
            image=self.get_icon(path))
        self.fsobjects[iid] = path
        return iid

    def load_tree(self, path: Path, parent: str = ""):
        """
        Load the contents of `path` into the treeview.
        """
        for fsobj in self.safe_iterdir(path):
            fullpath = path / fsobj
            child = self.insert_item(fsobj.name, fullpath, parent)
            # Preload the content of each directory within `path`.
            # This is necessary to make the folder item expandable.
            if fullpath.is_dir():
                for sub_fsobj in self.safe_iterdir(fullpath):
                    self.insert_item(sub_fsobj.name, fullpath/sub_fsobj, child)

    def load_subitems(self, iid: str):
        """
        Load the content of each folder inside the specified item
        into the treeview.
        """
        for child_iid in self.treeview.get_children(iid):
            if self.fsobjects[child_iid].is_dir():
                self.load_tree(self.fsobjects[child_iid],
                               parent=child_iid)

    def item_opened(self, _event: Event):
        """
        Handler invoked when a folder item is expanded.
        """
        # Get the expanded item.
        iid = self.treeview.selection()
        # If it is a folder, loads its content.
        self.load_subitems(iid)

    def hide(self):
        """ Removes treeview from interface """
        self.tv_pane.grid_remove()
        return

    def show(self):
        self.tv_pane.grid()
        return



"""
    FoxDot __main__.py
    ------------------

    Use FoxDot's interface by running this as a Python script, e.g.
    python __main__.py or python -m FoxDot if you have FoxDot correctly
    installed and Python on your path.

"""
from renardo_lib.runtime import FoxDotCode, handle_stdin
from FoxDotEditor.Editor import workspace
import argparse

parser = argparse.ArgumentParser(
    prog="Renardo",
    description="Live coding with Python and SuperCollider",
    epilog="More information: https://renardo.org/")

parser.add_argument('-d', '--dir',
                    action='store',
                    help="Use an alternate directory for looking up samples")
parser.add_argument('-s', '--startup',
                    action='store',
                    help="Use an alternate startup file")
parser.add_argument('-S', '--simple',
                    action='store_true',
                    help="Run Renardo in simple (accessible) mode")
parser.add_argument('-n', '--no-startup',
                    action='store_true',
                    help="Does not load startup.py on boot")
parser.add_argument('-b', '--boot',
                    action='store_true',
                    help="Boot SuperCollider from the command line")

args = parser.parse_args()

if args.dir:
    try:
        # Use given directory
        FoxDotCode.use_sample_directory(args.dir)
    except OSError as Exception:
        # Exit with last error
        import sys
        import traceback
        sys.exit(traceback.print_exc(limit=1))

if args.startup:
    try:
        FoxDotCode.use_startup_file(args.startup)
    except OSError as Exception:
        import sys
        import traceback
        sys.exit(traceback.print_exc(limit=1))

if args.no_startup:
    FoxDotCode.no_startup()

if args.boot:
    FoxDotCode.boot_supercollider()

# Open the GUI
FoxDot = workspace(FoxDotCode).run()




from FoxDotEditor.tkimport import *
from renardo_lib.Settings import *
from .Format import *
from ttkbootstrap.dialogs.colorchooser import ColorChooserDialog
try:
    import tkMessageBox
except ImportError:
    from tkinter import messagebox as tkMessageBox
import os.path


class Preferences:
    def __init__(self):
        self.w = 850
        self.h = 600
        self.stop = tb.Toplevel(topmost=True)
        self.stop.title("Preferences")
        self.stop.protocol("WM_DELETE_WINDOW", self.save_and_close)
        self.stop.minsize(400, 300)
        self.stop.resizable(True, True)
        self.stop.geometry(str(self.w)+"x"+str(self.h))
        try:
            # Use .ico file by default
            self.stop.iconbitmap(FOXDOT_ICON)
        except TclError:
            # Use .gif if necessary
            self.stop.tk.call('wm',
                              'iconphoto',
                              self.stop._w,
                              PhotoImage(file=FOXDOT_ICON_GIF))
        self.text_themes = ()
        for file_name in [file for file in os.listdir(FOXDOT_EDITOR_THEMES) if file.endswith('.json')]:
            theme = os.path.splitext(file_name)[0]
            self.text_themes = self.text_themes + (theme,)
        self.settings = {}
        self.conf_json = FOXDOT_CONFIG_FILE
        self.tabview = tb.Notebook(self.stop)
        self.general = tb.Frame(self.tabview)
        self.colors = tb.Frame(self.tabview)
        self.advanced = tb.Frame(self.tabview)
        self.tabview.add(self.general, text="General")  # add tab at the end
        self.tabview.add(self.colors, text="Appearance")
        self.tabview.add(self.advanced, text="Advanced")
        self.tabview.grid(row=0, column=0, columnspan=4, sticky="nsew")
        self.alert_text = "MODIFYING THIS FILE WILL OVERWRITE CHANGES DONE IN 'General' AND 'Appearance'. SAVE 'Preferences' FIRST TO CONTINUE!"
        self.alert = tb.Label(self.advanced,
                              text=self.alert_text)
        self.alert.grid(row=0, column=0, columnspan=3, padx=10, pady=20, sticky='nsew')
        self.y_scroll = tb.Scrollbar(self.advanced)
        self.y_scroll.grid(row=1, column=4, sticky='nsew')
        self.textbox = Text(self.advanced,
                            yscrollcommand=self.y_scroll.set)
        self.textbox.grid(row=1, column=0, columnspan=3, sticky='nsew')
        self.y_scroll.configure(command=self.textbox.yview)
        self.exit = tb.Button(
            self.stop,
            text="Cancel",
            command=self.save_and_close)
        self.exit.grid(row=1, column=2, padx=20, sticky="ew")
        self.save = tb.Button(
            self.stop,
            text="Save Changes",
            command=self.save_and_close)
        self.save.grid(row=1, column=3, padx=20, sticky="ew")
        self.unsaved = True
        try:
            with open(self.conf_json) as f:
                self.text = f.read().rstrip()
            self.textbox.insert(INSERT, self.text)
        except FileNotFoundError:
            print("conf.json file not found")
        # Add binds?
        self.textbox.bind()
        self.theme_name = ""
        # Settings values
        self.menu_start = BooleanVar()
        self.menu_start.set(MENU_ON_STARTUP)
        self.linenumbers_start = BooleanVar()
        self.linenumbers_start.set(LINENUMBERS_ON_STARTUP)
        self.console_start = BooleanVar()
        self.console_start.set(CONSOLE_ON_STARTUP)
        self.treeview_start = BooleanVar()
        self.treeview_start.set(TREEVIEW_ON_STARTUP)
        self.recover_work = BooleanVar()
        self.recover_work.set(RECOVER_WORK)
        self.check_update = BooleanVar()
        self.check_update.set(CHECK_FOR_UPDATE)
        # Editor
        self.linenumber_offset = StringVar()
        self.linenumber_offset.set(LINE_NUMBER_MARKER_OFFSET)
        self.brackets_auto = BooleanVar()
        self.brackets_auto.set(AUTO_COMPLETE_BRACKETS)
        # Sample values
        self.samples_dir = StringVar()
        self.samples_dir.set(SAMPLES_DIR)
        self.sample_pack = StringVar()
        self.sample_pack.set(SAMPLES_PACK_NUMBER)
        # Connection values
        self.address = StringVar()
        self.address.set(ADDRESS)
        self.port = StringVar()
        self.port.set(PORT)
        self.port2 = StringVar()
        self.port2.set(PORT2)
        self.fwd_address = StringVar()
        self.fwd_address.set(FORWARD_ADDRESS)
        self.fwd_port = StringVar()
        self.fwd_port.set(FORWARD_PORT)
        # SuperCollider
        self.sc_start = BooleanVar()
        self.sc_start.set(BOOT_ON_STARTUP)
        self.sc_path = StringVar()
        self.sc_path.set(SUPERCOLLIDER)
        self.sc3_start = BooleanVar()
        self.sc3_start.set(SC3_PLUGINS)
        self.max_ch = StringVar()
        self.max_ch.set(MAX_CHANNELS)
        self.sc_info = BooleanVar()
        self.sc_info.set(GET_SC_INFO)
        # Performance
        self.cpu_use = StringVar()
        cpu_val = self.convert2str(CPU_USAGE)
        self.cpu_use.set(cpu_val)
        lat_val = self.convert2str(CLOCK_LATENCY)
        self.clk_lat = StringVar()
        self.clk_lat.set(lat_val)
        # Appearance
        # theme
        self.theme = StringVar()
        self.theme.set(COLOR_THEME)
        self.text_theme = StringVar()
        self.text_theme.set(TEXT_COLORS)
        self.font = StringVar()
        self.font.set(FONT)
        self.use_alpha = BooleanVar()
        self.use_alpha.set(USE_ALPHA)
        self.alpha_val = StringVar()
        self.alpha_val.set(ALPHA_VALUE)
        self.alpha_start = BooleanVar()
        self.alpha_start.set(TRANSPARENT_ON_STARTUP)
        # Colors
        # ------------------
        self.plaintext = colour_map['plaintext']
        self.background = colour_map['background']
        self.functions = colour_map['functions']
        self.key_types = colour_map['key_types']
        self.user_defn = colour_map['user_defn']
        self.other_kws = colour_map['other_kws']
        self.comments = colour_map['comments']
        self.numbers = colour_map['numbers']
        self.strings = colour_map['strings']
        self.dollar = colour_map['dollar']
        self.arrow = colour_map['arrow']
        self.players = colour_map['players']
        # Prompt colours
        # ------------------
        self.prompt_fg = colour_map['prompt_fg']
        self.prompt_bg = colour_map['prompt_bg']
        # Console area colours
        # ------------------
        self.console_text = colour_map['console_text']
        self.console_bg = colour_map['console_bg']
        # x distance for widgets
        self.padx = 10
        # Widgets in section General
        # ON START
        self.g1 = tb.Frame(self.general)
        self.g1.grid(row=0, column=0, sticky="n")
        self.g2 = tb.Frame(self.general)
        self.g2.grid(row=0, column=1, sticky="n")
        self.g3 = tb.Frame(self.general)
        self.g3.grid(row=0, column=2, sticky="n")
        self.g4 = tb.Frame(self.general)
        self.g4.grid(row=0, column=3, sticky="n")
        self.lbl_on_start = tb.Label(
            self.g1,
            text="ACTIVATED ON START")
        self.lbl_on_start.grid(column=0, row=0, padx=self.padx*2, pady=10,
                               sticky="nw")
        self.tgl_menu = tb.Checkbutton(
            self.g1, text="Menu", variable=self.menu_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_menu.grid(column=0, row=1, padx=self.padx*2, pady=10,
                           sticky="nw")
        self.tgl_console = tb.Checkbutton(
            self.g1, text="Console", variable=self.console_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_console.grid(column=0, row=2, padx=self.padx*2, pady=10,
                              sticky="nw")
        self.tgl_linenumbers = tb.Checkbutton(
            self.g1, text="Line Numbers", variable=self.linenumbers_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_linenumbers.grid(column=0, row=2, padx=self.padx*2, pady=10,
                                  sticky="nw")
        self.tgl_treeview = tb.Checkbutton(
            self.g1, text="Treeview", variable=self.treeview_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_treeview.grid(column=0, row=3, padx=self.padx*2, pady=10,
                               sticky="nw")
        self.tgl_recover = tb.Checkbutton(
            self.g1, text="Recover Work", variable=self.recover_work,
            style='Roundtoggle.Toolbutton')
        self.tgl_recover.grid(column=0, row=4, padx=self.padx*2, pady=10,
                              sticky="nw")
        self.tgl_update = tb.Checkbutton(
            self.g1, text="Check for Updates", variable=self.check_update,
            style='Roundtoggle.Toolbutton')
        self.tgl_update.grid(column=0, row=5, padx=self.padx*2, pady=10,
                             sticky="nw")
        # EDITOR
        self.lbl_on_start = tb.Label(
            self.g1, text="EDITOR")
        self.lbl_on_start.grid(column=0, row=6, padx=self.padx*2, pady=10,
                               sticky="nw")
        self.tgl_update = tb.Checkbutton(
            self.g1, text="Autocomplete Brackets", variable=self.brackets_auto,
            style='Roundtoggle.Toolbutton')
        self.tgl_update.grid(column=0, row=7, padx=self.padx*2, pady=10,
                             sticky="nw")
        self.lbl_line_offset = tb.Label(
            self.g1, text="Linenumber Marker Offset")
        self.lbl_line_offset.grid(column=0, row=8, padx=self.padx*2,
                                  sticky="sw")
        self.entry_linenum_offset = tb.Entry(
            self.g1, textvariable=self.linenumber_offset)
        self.entry_linenum_offset.grid(column=0, row=9, padx=self.padx*2,
                                       pady=5, sticky="nw")
        # CONNECTIONS
        self.lbl_connect = tb.Label(
            self.g2, text="CONNECTIONS")
        self.lbl_connect.grid(column=1, row=0, padx=self.padx, pady=10,
                              sticky="nw")
        self.lbl_address = tb.Label(
            self.g2, text="Address")
        self.lbl_address.grid(column=1, row=1, padx=self.padx, sticky="sw")
        self.entry_address = tb.Entry(
            self.g2, textvariable=self.address)
        self.entry_address.grid(column=1, row=2, padx=self.padx, sticky="nw")
        self.lbl_port = tb.Label(
            self.g2, text="Port")
        self.lbl_port.grid(column=1, row=3, padx=self.padx, sticky="sw")
        self.entry_port = tb.Entry(
            self.g2, textvariable=self.port)
        self.entry_port.grid(column=1, row=4, padx=self.padx, sticky="nw")
        self.lbl_port2 = tb.Label(
            self.g2, text="Port2")
        self.lbl_port2.grid(column=1, row=5, padx=self.padx, sticky="sw")
        self.entry_port2 = tb.Entry(
            self.g2, textvariable=self.port2)
        self.entry_port2.grid(column=1, row=6, padx=self.padx, sticky="nw")
        self.lbl_fwd_address = tb.Label(
            self.g2, text="Forward Address")
        self.lbl_fwd_address.grid(column=1, row=7, padx=self.padx, sticky="sw")
        self.entry_fwd_address = tb.Entry(
            self.g2, textvariable=self.fwd_address)
        self.entry_fwd_address.grid(column=1, row=8, padx=self.padx, pady=5,
                                    sticky="nw")
        self.lbl_fwd_port = tb.Label(
            self.g2, text="Forward Port")
        self.lbl_fwd_port.grid(column=1, row=9, padx=self.padx, sticky="sw")
        self.entry_fwd_port = tb.Entry(
            self.g2, textvariable=self.fwd_port)
        self.entry_fwd_port.grid(column=1, row=10, padx=self.padx, pady=5,
                                 sticky="nw")
        # SUPERCOLLIDER
        self.lbl_sc = tb.Label(
            self.g3, text="SUPERCOLLIDER")
        self.lbl_sc.grid(column=2, row=0, padx=self.padx, pady=10, sticky="nw")
        self.tgl_sc_start = tb.Checkbutton(
            self.g3, text="Boot SC on Start", variable=self.sc_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_sc_start.grid(column=2, row=1, padx=self.padx, pady=10,
                               sticky="sw")
        self.lbl_sc_path = tb.Label(
            self.g3, text="SuperCollider Path")
        self.lbl_sc_path.grid(column=2, row=2, padx=self.padx, sticky="sw")
        self.entry_sc_path = tb.Entry(
            self.g3, textvariable=self.sc_path)
        self.entry_sc_path.grid(column=2, row=3, padx=self.padx, pady=5,
                                sticky="nw")
        self.tgl_sc3_start = tb.Checkbutton(
            self.g3, text="Use sc3-plugins", variable=self.sc3_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_sc3_start.grid(column=2, row=4, padx=self.padx, pady=10,
                                sticky="nw")
        self.lbl_max_ch = tb.Label(
            self.g3, text="Max. Channels")
        self.lbl_max_ch.grid(column=2, row=5, padx=self.padx*2, sticky="sw")
        self.entry_max_ch = tb.Entry(
            self.g3, textvariable=self.max_ch)
        self.entry_max_ch.grid(column=2, row=6, padx=self.padx, pady=5,
                               sticky="nw")
        self.tgl_sc_info = tb.Checkbutton(
            self.g3, text="Get SC Info", variable=self.sc_info,
            style='Roundtoggle.Toolbutton')
        self.tgl_sc_info.grid(column=2, row=7, padx=self.padx, pady=10,
                              sticky="nw")
        # PERFORMANCE
        self.lbl_perf = tb.Label(
            self.g4, text="PERFORMANCE")
        self.lbl_perf.grid(column=3, row=0, padx=self.padx, pady=10,
                           sticky="nw")
        self.lbl_cpu_use = tb.Label(
            self.g4, text="CPU Usage")
        self.lbl_cpu_use.grid(column=3, row=1, padx=self.padx, sticky="sw")
        self.cpu_use_opt = tb.Combobox(
            self.g4, textvariable=self.cpu_use)
        self.cpu_use_opt["values"] = ("Low", "Medium", "High")
        self.cpu_use_opt.grid(column=3, row=2, padx=self.padx, pady=10,
                              sticky="nw")
        self.lbl_clk_lat = tb.Label(
            self.g4, text="Clock Latency")
        self.lbl_clk_lat.grid(column=3, row=3, padx=self.padx, sticky="sw")
        self.clk_lat_opt = tb.Combobox(
            self.g4, textvariable=self.clk_lat)
        self.clk_lat_opt["values"] = ("Low", "Medium", "High")
        self.clk_lat_opt.grid(column=3, row=4, padx=self.padx, pady=10,
                              sticky="nw")
        # SAMPLES
        self.lbl_samples = tb.Label(
            self.g4, text="SAMPLES")
        self.lbl_samples.grid(column=3, row=5, padx=self.padx*2, pady=10,
                              sticky="nw")
        self.lbl_smpldir = tb.Label(
            self.g4, text="Samples Directory")
        self.lbl_smpldir.grid(column=3, row=6, padx=self.padx*2, sticky="sw")
        self.entry_smpldir = tb.Entry(
            self.g4, textvariable=self.samples_dir)
        self.entry_smpldir.grid(column=3, row=7, padx=self.padx*2, pady=5,
                                sticky="nw")
        self.lbl_smplpck = tb.Label(
            self.g4, text="Samplepack #")
        self.lbl_smplpck.grid(column=3, row=8, padx=self.padx*2,
                              sticky="nw")
        self.entry_smplpck = tb.Entry(
            self.g4, textvariable=self.sample_pack)
        self.entry_smplpck.grid(column=3, row=9, padx=self.padx*2, pady=5,
                                sticky="nw")
        # APPEARENCE COLOURS
        self.a1 = tb.Frame(self.colors)
        self.a1.grid(row=0, column=0, padx=self.padx*2, sticky="nw")
        self.a2 = tb.Frame(self.colors)
        self.a2.grid(row=0, column=1, padx=self.padx*2, sticky="nw")
        self.a3 = tb.Frame(self.a2)
        self.a3.grid(row=0, column=0, sticky="nw")
        self.a4 = tb.Frame(self.a2)
        self.a4.grid(row=1, column=0, pady=self.padx, sticky="nw")
        self.a5 = tb.Frame(self.a2)
        self.a5.grid(row=1, column=1, pady=self.padx, sticky="nw")
        self.lbl_themes = tb.Label(
            self.a1, text="Theme")
        self.lbl_themes.grid(column=0, row=0, padx=self.padx, pady=self.padx/2,
                             sticky="nw")
        self.themes_opt = tb.Combobox(
            self.a1, textvariable=self.theme)
        self.themes_opt["values"] = ("cosmo", "flatly",
                                     "journal", "litera", "lumen", "minty",
                                     "pulse", "sandstone", "united", "yeti",
                                     "morph", "simplex", "cerulean", "solar",
                                     "superhero", "darkly", "cyborg", "vapor")
        self.themes_opt.grid(column=0, row=1, padx=self.padx, pady=self.padx/2,
                             sticky="nw")
        self.lbl_text_colors = tb.Label(
            self.a1, text="Text Color Theme")
        self.lbl_text_colors.grid(column=0, row=2, padx=self.padx,
                                  pady=self.padx, sticky="sw")
        self.text_colors_opt = tb.Combobox(
            self.a1, textvariable=self.text_theme)
        self.text_colors_opt["values"] = self.text_themes
        self.text_colors_opt.grid(column=0, row=3, padx=self.padx,
                                  pady=self.padx/2, sticky="nw")
        self.btn_load = tb.Button(self.a1, text="Load",
                                  command=self.load_tt)
        self.btn_load.grid(column=0, row=4, padx=self.padx*3,
                           pady=self.padx/2, sticky="sw")
        self.btn_save = tb.Button(self.a1, text="Save", command=self.save_tt)
        self.btn_save.grid(column=0, row=4, padx=self.padx*12,
                           pady=self.padx/2, sticky="sw")
        self.lbl_font = tb.Label(
            self.a1, text="Font")
        self.lbl_font.grid(column=0, row=5, padx=self.padx,
                           pady=self.padx/2, sticky="sw")
        self.entry_font = tb.Entry(
            self.a1, textvariable=self.font)
        self.entry_font.grid(column=0, row=6, padx=self.padx, pady=self.padx/2,
                             sticky="nw")
        self.tgl_use_alpha = tb.Checkbutton(
            self.a1, text="Use Alpha", variable=self.use_alpha,
            style='Roundtoggle.Toolbutton')
        self.tgl_use_alpha.grid(column=0, row=7, padx=self.padx,
                                pady=self.padx, sticky="nw")
        self.tgl_alpha_start = tb.Checkbutton(
            self.a1, text="Transparent on Start", variable=self.alpha_start,
            style='Roundtoggle.Toolbutton')
        self.tgl_alpha_start.grid(column=0, row=8, padx=self.padx,
                                  pady=self.padx, sticky="nw")
        # COLOR CHOICE
        self.lbl_colorlist = tb.Label(self.a3, text="Text Colors Editor")
        self.lbl_colorlist.grid(column=0, row=0,
                                padx=self.padx,
                                pady=self.padx,
                                sticky="nw")
        # plaintext
        self.lbl_plaintext = tb.Label(
            self.a4, text="plaintext")
        self.lbl_plaintext.grid(column=0, row=0, padx=self.padx, sticky="nw")
        self.btn_plaintext = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_plaintext, self.plaintext))
        self.btn_plaintext.config(bg=self.plaintext)
        self.btn_plaintext.grid(column=1, row=0, sticky="nw")
        # background
        self.lbl_background = tb.Label(
            self.a4, text="background")
        self.lbl_background.grid(column=0, row=2, padx=self.padx, sticky="nw")
        self.btn_background = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_background, self.background))
        self.btn_background.config(bg=self.background)
        self.btn_background.grid(column=1, row=2, sticky="nw")
        # functions
        self.lbl_functions = tb.Label(
            self.a4, text="functions")
        self.lbl_functions.grid(column=0, row=3, padx=self.padx, sticky="nw")
        self.btn_functions = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_functions, self.functions))
        self.btn_functions.config(bg=self.functions)
        self.btn_functions.grid(column=1, row=3, sticky="nw")
        # key_types
        self.lbl_key_types = tb.Label(
            self.a4, text="key_types")
        self.lbl_key_types.grid(column=0, row=4, padx=self.padx, sticky="nw")
        self.btn_key_types = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_key_types, self.key_types))
        self.btn_key_types.config(bg=self.key_types)
        self.btn_key_types.grid(column=1, row=4, sticky="nw")
        # user_defn
        self.lbl_user_defn = tb.Label(
            self.a4, text="user_defn")
        self.lbl_user_defn.grid(column=0, row=5, padx=self.padx, sticky="nw")
        self.btn_user_defn = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_user_defn, self.user_defn))
        self.btn_user_defn.config(bg=self.user_defn)
        self.btn_user_defn.grid(column=1, row=5, sticky="nw")
        # other_kws
        self.lbl_other_kws = tb.Label(
            self.a4, text="other_kws")
        self.lbl_other_kws.grid(column=0, row=6, padx=self.padx, sticky="nw")
        self.btn_other_kws = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_other_kws, self.other_kws))
        self.btn_other_kws.config(bg=self.other_kws)
        self.btn_other_kws.grid(column=1, row=6, sticky="nw")
        # comments
        self.lbl_comments = tb.Label(
            self.a4, text="comments")
        self.lbl_comments.grid(column=0, row=7, padx=self.padx, sticky="nw")
        self.btn_comments = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_comments, self.comments))
        self.btn_comments.config(bg=self.comments)
        self.btn_comments.grid(column=1, row=7, sticky="nw")
        # numbers
        self.lbl_numbers = tb.Label(
            self.a4, text="numbers")
        self.lbl_numbers.grid(column=0, row=8, padx=self.padx, sticky="nw")
        self.btn_numbers = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_numbers, self.numbers))
        self.btn_numbers.config(bg=self.numbers)
        self.btn_numbers.grid(column=1, row=8, sticky="nw")
        # strings
        self.lbl_strings = tb.Label(
            self.a4, text="strings")
        self.lbl_strings.grid(column=0, row=9, padx=self.padx, sticky="nw")
        self.btn_strings = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_strings, self.strings))
        self.btn_strings.config(bg=self.strings)
        self.btn_strings.grid(column=1, row=9, sticky="nw")
        # dollar
        self.lbl_dollar = tb.Label(
            self.a4, text="dollar")
        self.lbl_dollar.grid(column=0, row=10, padx=self.padx, sticky="nw")
        self.btn_dollar = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_dollar, self.dollar))
        self.btn_dollar.config(bg=self.dollar)
        self.btn_dollar.grid(column=1, row=10, sticky="nw")
        # arrow
        self.lbl_arrow = tb.Label(
            self.a4, text="arrow")
        self.lbl_arrow.grid(column=0, row=11, padx=self.padx, sticky="nw")
        self.btn_arrow = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_arrow, self.arrow))
        self.btn_arrow.config(bg=self.arrow)
        self.btn_arrow.grid(column=1, row=11, sticky="nw")
        # players
        self.lbl_players = tb.Label(
            self.a4, text="players")
        self.lbl_players.grid(column=0, row=12, padx=self.padx, sticky="nw")
        self.btn_players = Button(
            self.a4, width=5,
            command=lambda: self.cc(self.btn_players, self.players))
        self.btn_players.config(bg=self.players)
        self.btn_players.grid(column=1, row=12, sticky="nw")
        # prompt_fg
        self.lbl_prompt_fg = tb.Label(
            self.a5, text="prompt_fg")
        self.lbl_prompt_fg.grid(column=0, row=0, padx=self.padx, sticky="nw")
        self.btn_prompt_fg = Button(
            self.a5, width=5,
            command=lambda: self.cc(self.btn_prompt_fg, self.prompt_fg))
        self.btn_prompt_fg.config(bg=self.prompt_fg)
        self.btn_prompt_fg.grid(column=1, row=0, sticky="nw")
        # prompt_bg
        self.lbl_prompt_bg = tb.Label(
            self.a5, text="prompt_bg")
        self.lbl_prompt_bg.grid(column=0, row=1, padx=self.padx, sticky="nw")
        self.btn_prompt_bg = Button(
            self.a5, width=5,
            command=lambda: self.cc(self.btn_prompt_bg, self.prompt_bg))
        self.btn_prompt_bg.config(bg=self.prompt_bg)
        self.btn_prompt_bg.grid(column=1, row=1, sticky="nw")
        # console_text
        self.lbl_console_text = tb.Label(
            self.a5, text="console_text")
        self.lbl_console_text.grid(column=0, row=2, padx=self.padx, sticky="nw")
        self.btn_console_text = Button(
            self.a5, width=5,
            command=lambda: self.cc(self.btn_console_text, self.console_text))
        self.btn_console_text.config(bg=self.console_text)
        self.btn_console_text.grid(column=1, row=2, sticky="nw")
        # console_bg
        self.lbl_console_bg = tb.Label(
            self.a5, text="console_bg")
        self.lbl_console_bg.grid(column=0, row=3, padx=self.padx, sticky="nw")
        self.btn_console_bg = Button(
            self.a5, width=5,
            command=lambda: self.cc(self.btn_console_bg, self.console_bg))
        self.btn_console_bg.config(bg=self.console_bg)
        self.btn_console_bg.grid(column=1, row=3, sticky="nw")

    def load_tt(self):
        self.theme_name = self.text_colors_opt.get()
        try:
            file = FOXDOT_EDITOR_THEMES + '/' + self.theme_name + '.json'
            # Opening JSON file
            with open(file, 'r') as openfile:
                # Reading from json file
                json_object = json.load(openfile)
                self.btn_plaintext.config(
                    bg=json_object[self.theme_name]['plaintext'])
                self.btn_background.config(
                    bg=json_object[self.theme_name]['background'])
                self.btn_functions.config(
                    bg=json_object[self.theme_name]['functions'])
                self.btn_key_types.config(
                    bg=json_object[self.theme_name]['key_types'])
                self.btn_user_defn.config(
                    bg=json_object[self.theme_name]['user_defn'])
                self.btn_other_kws.config(
                    bg=json_object[self.theme_name]['other_kws'])
                self.btn_comments.config(
                    bg=json_object[self.theme_name]['comments'])
                self.btn_numbers.config(
                    bg=json_object[self.theme_name]['numbers'])
                self.btn_strings.config(
                    bg=json_object[self.theme_name]['strings'])
                self.btn_dollar.config(
                    bg=json_object[self.theme_name]['dollar'])
                self.btn_arrow.config(
                    bg=json_object[self.theme_name]['arrow'])
                self.btn_players.config(
                    bg=json_object[self.theme_name]['players'])
                self.btn_prompt_fg.config(
                    bg=json_object[self.theme_name]['prompt_fg'])
                self.btn_prompt_bg.config(
                    bg=json_object[self.theme_name]['prompt_bg'])
                self.btn_console_text.config(
                    bg=json_object[self.theme_name]['console_text'])
                self.btn_console_bg.config(
                    bg=json_object[self.theme_name]['console_bg'])
        except FileNotFoundError:
            pass

    def save_tt(self):
        # Colors
        # ------------------
        new_theme = {self.theme_name: {}}
        new_theme[self.theme_name]['plaintext'] = self.plaintext
        new_theme[self.theme_name]['background'] = self.background
        new_theme[self.theme_name]['functions'] = self.functions
        new_theme[self.theme_name]['key_types'] = self.key_types
        new_theme[self.theme_name]['user_defn'] = self.user_defn
        new_theme[self.theme_name]['other_kws'] = self.other_kws
        new_theme[self.theme_name]['comments'] = self.comments
        new_theme[self.theme_name]['numbers'] = self.numbers
        new_theme[self.theme_name]['strings'] = self.strings
        new_theme[self.theme_name]['dollar'] = self.dollar
        new_theme[self.theme_name]['arrow'] = self.arrow
        new_theme[self.theme_name]['players'] = self.players
        # Prompt colours
        # ------------------
        new_theme[self.theme_name]['prompt_fg'] = self.prompt_fg
        new_theme[self.theme_name]['prompt_bg'] = self.prompt_bg
        # Console area colours
        # ------------------
        new_theme[self.theme_name]['console_text'] = self.console_text
        new_theme[self.theme_name]['console_bg'] = self.console_bg
        self.stop.iconify()
        self.filename = tkFileDialog.asksaveasfilename(
            filetypes=[("JSON files", ".json")],
            initialdir=FOXDOT_EDITOR_THEMES + '/',
            defaultextension=".json")
        if self.filename:
            new_file = open(self.filename, "w")
            json.dump(new_theme, new_file, indent=6)
            new_file.close()
            print("Theme saved.")
        else:
            pass
        self.stop.deiconify()

    def cc(self, button, color):
        self.cchooser = ColorChooserDialog(initialcolor=color)
        self.stop.iconify()
        self.cchooser.show()
        self.stop.deiconify()
        if not self.cchooser.result:
            pass
        else:
            colors = self.cchooser.result
            color = colors.hex
            button.config(bg=color)

    def convert2number(self, selection, option):
        if selection == "Low":
            if option == 0:
                self.cpu_use.set("0")
            if option == 1:
                self.clk_lat.set("0")
        elif selection == "Medium":
            if option == 0:
                self.cpu_use.set("1")
            if option == 1:
                self.clk_lat.set("1")
        elif selection == "High":
            if option == 0:
                self.cpu_use.set("2")
            if option == 1:
                self.clk_lat.set("2")

    def convert2str(self, variable):
        if variable == 0:
            return "Low"
        elif variable == 1:
            return "Medium"
        elif variable == 2:
            return "High"

    def start(self):
        self.stop.mainloop()

    def save_and_close(self, event=None):
        """ Asks the user if they want to save changes """
        answer = tkMessageBox.askyesno("Save changes",
                                       "Do you want to save your changes?",
                                       parent=self.stop)
        self.stop.lift(aboveThis=None)
        if answer:
            return self.save_changes()
        else:
            pass
        return self.stop.destroy()
        pass

    def save_changes(self):
        self.text_settings = self.textbox.get("1.0", "end-1c")
        if self.text_settings != self.text:
            try:
                self.settings.clear()
                self.settings = json.loads(self.text_settings)
            except Exception:
                print("Can not convert text to json file. Please check your changes in this file!")
        else:
            self.settings.clear()
            self.settings['ADDRESS'] = self.address.get()
            self.settings['PORT'] = int(self.port.get())
            self.settings['PORT2'] = int(self.port2.get())
            self.settings['FONT'] = self.font.get()
            self.settings['SUPERCOLLIDER'] = self.sc_path.get()
            self.settings['BOOT_ON_STARTUP'] = self.sc_start.get()
            self.settings['SC3_PLUGINS'] = self.sc3_start.get()
            self.settings['MAX_CHANNELS'] = int(self.max_ch.get())
            self.settings['SAMPLES_DIR'] = self.samples_dir.get()
            self.settings['SAMPLES_PACK_NUMBER'] = int(self.sample_pack.get())
            self.settings['GET_SC_INFO'] = self.sc_info.get()
            self.settings['USE_ALPHA'] = self.use_alpha.get()
            self.settings['ALPHA_VALUE'] = float(self.alpha_val.get())
            self.settings['MENU_ON_STARTUP'] = self.menu_start.get()
            self.settings['CONSOLE_ON_STARTUP'] = self.console_start.get()
            self.settings['LINENUMBERS_ON_STARTUP'] = self.linenumbers_start.get()
            self.settings['TREEVIEW_ON_STARTUP'] = self.treeview_start.get()
            self.settings['TRANSPARENT_ON_STARTUP'] = self.alpha_start.get()
            self.settings['RECOVER_WORK'] = self.recover_work.get()
            self.settings['CHECK_FOR_UPDATE'] = self.check_update.get()
            self.settings['LINE_NUMBER_MARKER_OFFSET'] = int(self.linenumber_offset.get())
            self.settings['AUTO_COMPLETE_BRACKETS'] = self.brackets_auto.get()
            self.convert2number(self.cpu_use.get(), 0)
            self.settings['CPU_USAGE'] = int(self.cpu_use.get())
            self.convert2number(self.clk_lat.get(), 1)
            self.settings['CLOCK_LATENCY'] = int(self.clk_lat.get())
            self.settings['FORWARD_ADDRESS'] = self.fwd_address.get()
            self.settings['FORWARD_PORT'] = int(self.fwd_port.get())
            self.settings['COLOR_THEME'] = self.theme.get()
            self.settings['TEXT_COLORS'] = self.text_theme.get()
        settings_file = open(self.conf_json, "w")
        json.dump(self.settings, settings_file, indent=6)
        settings_file.close()
        msg = "A restart of FoxDot is required for the changes to take effect"
        tkMessageBox.showwarning(parent=self.stop, title="Just a heads up", message=msg)
        self.stop.destroy()
        return




# -*- coding: utf-8 -*-
import re
from os.path import abspath, join, dirname
from renardo_lib.Code import classes, functions, instances
from renardo_lib.Settings import *
from renardo_lib.Patterns import Main, Sequences, Generators
from renardo_lib.SynthDefManagement import SCLangExperimentalPythonBindings


def path(fn):
    return abspath(join(dirname(__file__), fn))


# RegEx Group 1
user_defn = r"(?<=def )(\w+)"
players = r"(?<=>>)(\s*\w+)"
comments = r"^\s*#.*|[^\"']*(#[^\"']*$)"
decorator = r"\A\s*@.+"
numbers = r"\W+(\d+)"
strings = r"\".*?\"|\".*" + "|\'.*?\'|\'.*"
dollar = r"\s\$\s?"
arrow = r"\s>>\s?"
multiline = r"\"\"\".*\"\"\"|'''.*'''"


# RegEx Group 2
def re_list(string, br="[]"):
    new = ""
    nested = False
    n, m = 0, 0
    while n < len(string):
        char = string[n]
        if char == br[0]:
            if nested is False:
                nested = True
                m = n
                n += 1
            else:
                sub = string[n + 1:string.rfind(br[1])]
                value = re_list(sub)
                n += len(value) + 1
        elif char == br[1]:
            new_string = string[m:n+1]
            return new_string
        else:
            n += 1


def re_pat(string):
    r = r"P?" + re_list(string)
    for char in "[]()":
        r = r.replace(char, "\\" + char)
    return re.search(r, string)


def findstyles(line, *args):
    """ Finds any locations of any regex and returns the name
        of the style and the start and end point in the line """

    if not args:
        tags = re_patterns
    else:
        tags = [a for a in args if a in re_patterns]

    pos = []

    for tag in tags:
        match_start = match_end = 0
        for match in re.finditer(re_patterns[tag], line):
            looking = True
            i = 0
            while looking:
                try:
                    start = match.start(i)
                    end = match.end(i)
                    if start == end == -1:
                        raise IndexError  # this is hacky af
                    match_start = start
                    match_end = end

                except IndexError:
                    looking = False
                i += 1
            pos.append((tag, match_start, match_end))
    return pos


def userdefined(line):
    """ Returns the name of a user defined function or class in a string
        i.e. def foo(x) -> foo """
    match = re.search(user_defn, line)
    if match:
        return match.group()


def find_comment(line):
    """ Finds the index of a comment # and returns None if not found """
    instring, instring_char = False, ""
    for i, char in enumerate(line):
        if char in ('"', "'"):
            if instring:
                if char == instring_char:
                    instring = False
                    instring_char = ""
            else:
                instring = True
                instring_char = char
        elif char == "#":
            if not instring:
                return i
    return None


def find_multiline(text):
    pos = []
    tag = multiline
    match_start = match_end = 0
    for match in re.finditer(tag, text, re.DOTALL):
        looking = True
        i = 0
        while looking:
            try:
                start = match.start(i)
                end = match.end(i)
                if start == end == -1:
                    raise IndexError  # this is hacky af
                match_start = start
                match_end = end
            except IndexError:
                looking = False
            i += 1
        pos.append((match_start, match_end))
    return pos


# Use our regex to read patterns.py and add all the functions to key_types
foxdot_kw = ["Clock", "Group", "Scale", "Server", "Root", "Samples", "var",
             "Pvar", "Attributes", "PatternMethods", "PatternTypes", "now",
             "linvar", "expvar", "mapvar", "sinvar", "inf", "when", "lambda",
             u"λ", decorator]

foxdot_funcs = classes(Main) + classes(Generators) + functions(Sequences) + ["P"]

# Python keywords used in RegEx Group 2
py_indent_kw = ["for", "if", "elif", "else", "def", "while", "class", "try",
                "except"]

py_functions = ["if", "elif", "else", "return", "def", "print", "when", "from",
                "and", "or", "not", "is", "in", "for", "as", "with", "global",
                "while", "class", "import", "try", "except", "from", "break"]

py_other_kws = foxdot_kw + ["True", "False", "None", "self"]

py_key_types = ["abs", "divmod", "input", "open", "staticmethod", "all",
                "enumerate", "int", "ord", "str", "any", "eval", "isinstance",
                "pow", "sum", "basestring", "execfile", "issubclass", "super",
                "bin", "file", "iter", "property", "tuple", "bool", "filter",
                "len", "range", "type", "bytearray", "float", "list",
                "raw_input", "unichr", "callable", "format", "locals",
                "reduce", "unicode", "chr", "frozenset", "long", "reload",
                "vars", "classmethod", "getattr", "map", "repr", "xrange",
                "cmp", "globals", "max", "reversed", "zip", "compile",
                "hasattr", "memoryview", "round", "__import__", "complex",
                "hash", "min", "delattr", "set", "help", "next", "setattr",
                "dict", "hex", "object", "slice", "dir", "id", "sorted"]

py_key_types = foxdot_funcs + py_key_types
py_separators = list("[](){},./*+=- \t\n")
py_whitespace = list(" \t\n\r\f\v")
left_b = list("([{'\"")
right_b = list(")]}'\"")
brackets = dict([(right_b[i], left_b[i]) for i in range(3)])
tabsize = 4

# RegEx Group 2
functions = r"(?<![a-zA-Z._])(" + "|".join(py_functions) + ")(?![_a-zA-Z])"
key_types = r"(?<![a-zA-Z._])(" + "|".join(py_key_types) + ")(?![_a-zA-Z])"
other_kws = r"(?<![a-zA-Z._])(" + "|".join(py_other_kws) + ")(?![_a-zA-Z])"

colour_map = {'plaintext': COLOURS.plaintext,
              'background': COLOURS.background,
              'functions': COLOURS.functions,
              'key_types': COLOURS.key_types,
              'user_defn': COLOURS.user_defn,
              'other_kws': COLOURS.other_kws,
              'comments': COLOURS.comments,
              'numbers': COLOURS.numbers,
              'strings': COLOURS.strings,
              'dollar': COLOURS.dollar,
              'arrow': COLOURS.arrow,
              'players': COLOURS.players,
              'prompt_fg': COLOURS.prompt_fg,
              'prompt_bg': COLOURS.prompt_bg,
              'console_text': COLOURS.console_text,
              'console_bg': COLOURS.console_bg,
              'kick': COLOURS.kick,
              'various': COLOURS.various,
              'vocal': COLOURS.vocal,
              'bell': COLOURS.bell,
              'hihat': COLOURS.hihat,
              'clap': COLOURS.clap,
              'snap': COLOURS.snap,
              'shaker': COLOURS.shaker,
              'tambourine': COLOURS.tambourine,
              'crash': COLOURS.crash,
              'cymbal': COLOURS.cymbal,
              'soundfx': COLOURS.soundfx,
              'tom': COLOURS.tom,
              'noise': COLOURS.noise,
              'ride': COLOURS.ride,
              'perc': COLOURS.perc,
              'snare': COLOURS.snare,
              'rim': COLOURS.rim,
              'loops': COLOURS.loops,
              'default': COLOURS.default,
              'text1': COLOURS.text1,
              'text2': COLOURS.text2
              }

re_patterns = {'functions': functions,
               'key_types': key_types,
               'user_defn': user_defn,
               'other_kws': other_kws,
               'numbers': numbers,
               'strings': strings,
               'dollar': dollar,
               'arrow': arrow,
               'players': players
               }

bracket_formatting = {'borderwidth': 2,
                      'relief': 'groove'}

# Weightings (heavier get checked last)
tag_weights = [['numbers'],
               ['key_types',
                'functions',
                'user_defn',
                'other_kws',
                'dollar',
                'arrow',
                'players'],
               ['strings'],
               ['comments']]


def get_keywords():
    return list(set(py_indent_kw + py_functions + py_other_kws + py_key_types))



#!/usr/bin/python
""" Collection of functions used in the Workspace package """

from FoxDotEditor.Format import *
import sys


def stdout(*args):
    """ Forces prints to stdout and not console """
    sys.__stdout__.write(" ".join([str(s) for s in args]) + "\n")


def index(index1, index2=None):
    if type(index1) == str and index2 is None:
        return tuple(int(n) for n in index1.split('.'))
    return str(index1) + '.' + str(index2)


def empty(string):
    return len(string.replace(' ', '')) == 0


def isDelete(char):
    return char == "\x08"


def isReturn(char):
    return char == "\r"


def isHex(char):
    return len(repr(char)) > 3


def indented(s):
    if len(s) == 0:
        return 0
    for i, char in enumerate(s):
        if char != " ":
            break
    return i


def at_function(index, string):
    """ Returns true if the index is at a colon in the string """
    return False if index == 0 else string[index-1] == ":"


def in_brackets(index, string):
    """
    Returns true if index is between a pair of brackets (could be non closing)
    """
    left_b = list("({[")
    right_b = list(")}]")
    count = dict([(lst, 0) for lst in left_b])
    index = index - 1  # Look to the character to the left to start
    while index > 0:
        char = string[index]
        if char in left_b:
            if count[char] == 0:
                return True
            else:
                count[char] -= 1
        elif char in right_b:
            count[brackets[char]] += 1
        index -= 1
    return False


def get_tabspace(s):
    """ Returns the amount of whitespace at the start of a string """
    return indented(s) * " "


def open_bracket(text):
    """ Returns the index of the last open bracket """
    count = dict([(lst, 0) for lst in left_b])
    for i, char in enumerate(text):
        if char in left_b:
            count[char] += 1
        elif char in right_b:
            count[brackets[char]] -= 1
    open_br = [b for b in count.keys() if (count[b] >= max(count.values()) > 0) and (count[b] % 2 == 1)]
    pos = [text.rfind(b) for b in open_br]
    if pos:
        return max(pos) + 1


def function(text):
    tokens = text.strip()
    if tokens[-1] != ":":
        return
    tokens = tokens[:-1]
    for b in left_b + right_b:
        tokens = tokens.replace(b, " ")
    tokens = tokens.split()
    for kw in py_indent_kw:
        if kw in tokens:
            return indented(text) + tabsize



from FoxDotEditor.tkimport import Menu, BooleanVar, IntVar, DISABLED
from FoxDotEditor.tkimport import *
import os.path
from functools import partial
from renardo_lib.Settings import *
from renardo_lib.Code import FoxDotCode

# Code menu
ctrl = "Command" if SYSTEM == MAC_OS else "Ctrl"


class MenuBar(Menu):
    def __init__(self, master, visible=True):
        self.root = master
        self.menu_fontsize = ("", 11)
        Menu.__init__(self, master.root)
        # "ticked" menu options
        self.sc3_plugins = BooleanVar()
        self.sc3_plugins.set(SC3_PLUGINS)
        self.cpu_usage = IntVar()
        self.cpu_usage.set(CPU_USAGE)
        self.latency = IntVar()
        self.latency.set(CLOCK_LATENCY)
        # File menu
        self.filemenu = Menu(self, tearoff=0)
        self.filemenu.add_command(label="New Document",
                                  command=self.root.newfile,
                                  accelerator="Ctrl+N",
                                  font=self.menu_fontsize)
        self.filemenu.add_command(label="Open",
                                  command=self.root.openfile,
                                  accelerator="Ctrl+O",
                                  font=self.menu_fontsize)
        self.filemenu.add_command(label="Save",
                                  command=self.root.save,
                                  accelerator="Ctrl+S",
                                  font=self.menu_fontsize)
        self.filemenu.add_command(label="Save As...",
                                  command=self.root.saveAs,
                                  font=self.menu_fontsize)
        self.add_cascade(label="File",
                         menu=self.filemenu,
                         font=self.menu_fontsize)
        # Edit menu
        self.editmenu = Menu(self, tearoff=0)
        self.editmenu.add_command(label="Undo",
                                  command=self.root.undo,
                                  accelerator="Ctrl+Z",
                                  font=self.menu_fontsize)
        self.editmenu.add_command(label="Redo",
                                  command=self.root.redo,
                                  accelerator="Ctrl+Y",
                                  font=self.menu_fontsize)
        self.editmenu.add_separator()
        self.editmenu.add_command(label="Cut",
                                  command=self.root.edit_cut,
                                  accelerator="Ctrl+X",
                                  font=self.menu_fontsize)
        self.editmenu.add_command(label="Copy",
                                  command=self.root.edit_copy,
                                  accelerator="Ctrl+C",
                                  font=self.menu_fontsize)
        self.editmenu.add_command(label="Paste",
                                  command=self.root.edit_paste,
                                  accelerator="Ctrl+V",
                                  font=self.menu_fontsize)
        self.editmenu.add_command(label="Select All",
                                  command=self.root.select_all,
                                  accelerator="Ctrl+A",
                                  font=self.menu_fontsize)
        self.editmenu.add_separator()
        self.editmenu.add_command(label="Increase Font Size",
                                  command=self.root.zoom_in,
                                  accelerator="Ctrl+=",
                                  font=self.menu_fontsize)
        self.editmenu.add_command(label="Decrease Font Size",
                                  command=self.root.zoom_out,
                                  accelerator="Ctrl+-",
                                  font=self.menu_fontsize)
        self.editmenu.add_separator()
        self.editmenu.add_command(label="Preferences",
                                  accelerator="Ctrl+P",
                                  command=self.root.open_preferences,
                                  font=self.menu_fontsize)
        self.add_cascade(label="Edit",
                         menu=self.editmenu,
                         font=self.menu_fontsize)
        # Toolbars
        self.viewmenu = Menu(self, tearoff=0)
        if self.root.menu_toggled.get() is True:
            lbl_menu = "Hide Menu"
        else:
            lbl_menu = "Show Menu"
        self.viewmenu.add_command(label=lbl_menu,
                                  command=self.root.toggle_menu,
                                  accelerator="Ctrl+M",
                                  font=self.menu_fontsize)
        if self.root.linenumbers_toggled.get() is True:
            lbl_linenumbers = "Hide Line Numbers"
        else:
            lbl_linenumbers = "Show Line Numbers"
        self.viewmenu.add_command(label=lbl_linenumbers,
                                  command=self.root.toggle_linenumbers,
                                  accelerator="Ctrl+0",
                                  font=self.menu_fontsize)
        if self.root.treeview_toggled.get() is True:
            lbl_treeview = "Hide Treeview"
        else:
            lbl_treeview = "Show Treeview"
        self.viewmenu.add_command(label=lbl_treeview,
                                  command=self.root.toggle_treeview,
                                  accelerator="Ctrl+U",
                                  font=self.menu_fontsize)
        if self.root.searchbar_toggled.get() is True:
            lbl_searchbar = "Hide Searchbar"
        else:
            lbl_searchbar = "Show Searchbar"
        self.viewmenu.add_command(label=lbl_searchbar,
                                  command=self.root.toggle_searchbar,
                                  accelerator="Ctrl+F",
                                  font=self.menu_fontsize)
        self.viewmenu.add_separator()
        if self.root.console_toggled.get() is True:
            lbl_console = "Hide Console"
        else:
            lbl_console = "Show Console"
        self.viewmenu.add_command(label=lbl_console,
                                  command=self.root.toggle_console,
                                  font=self.menu_fontsize)
        self.viewmenu.add_command(label="Clear Console",
                                  command=self.root.clear_console,
                                  font=self.menu_fontsize)
        self.viewmenu.add_command(label="Export Console Log",
                                  command=self.root.export_console,
                                  font=self.menu_fontsize)
        self.viewmenu.add_separator()
        self.viewmenu.add_checkbutton(label="Toggle Fullscreen",
                                      command=(lambda: self.root.toggle_fullscreen(zoom=True)),
                                      variable=self.root.fullscreen_toggled,
                                      font=self.menu_fontsize)
        self.viewmenu.add_checkbutton(label="Toggle Window Transparency",
                                      command=self.root.toggle_transparency,
                                      variable=self.root.transparent,
                                      font=self.menu_fontsize)
        self.viewmenu.add_checkbutton(label="Toggle Beat Counter",
                                      command=self.root.toggle_counter,
                                      variable=self.root.show_counter,
                                      font=self.menu_fontsize)
        self.add_cascade(label="View",
                         menu=self.viewmenu,
                         font=self.menu_fontsize)
        self.codemenu = Menu(self, tearoff=0)
        self.codemenu.add_command(label="Evaluate Block",
                                  command=self.root.exec_block,
                                  accelerator="{}+Return".format(ctrl),
                                  font=self.menu_fontsize)
        self.codemenu.add_command(label="Evaluate Line",
                                  command=self.root.exec_line,
                                  accelerator="Alt+Return",
                                  font=self.menu_fontsize)
        self.codemenu.add_command(label="Clear Scheduling Clock",
                                  command=self.root.killall,
                                  accelerator="{}+.".format(ctrl),
                                  font=self.menu_fontsize)
        self.codemenu.add_separator()
        self.codemenu.add_checkbutton(label="Listen for connections",
                                      command=self.root.allow_connections,
                                      variable=self.root.listening_for_connections,
                                      font=self.menu_fontsize)
        self.add_cascade(label="Language",
                         menu=self.codemenu,
                         font=self.menu_fontsize)
        # Help
        self.helpmenu = Menu(self, tearoff=0)
        self.helpmenu.add_command(label="Display help message",
                                  command=self.root.help,
                                  accelerator="{}+{}".format(ctrl, self.root.help_key),
                                  font=self.menu_fontsize)
        self.helpmenu.add_command(label="Visit Renardo Homepage",
                                  command=self.root.openhomepage,
                                  font=self.menu_fontsize)
        self.helpmenu.add_command(label="Documentation",
                                  command=self.root.opendocumentation,
                                  font=self.menu_fontsize)
        # Tutorials
        self.tutomenu = Menu(self, tearoff=0)
        for tutorial in GET_TUTORIAL_FILES():
            filename = os.path.basename(tutorial)
            if filename.endswith(".py"):
                filename = filename.replace(".py", "")
                data = filename.split("_")
                num = data[0]
                name = " ".join(data[1:]).title()
                self.tutomenu.add_command(label="Load Tutorial {}: {}".format(num, name),
                                          command=partial(self.root.loadfile, tutorial))
        self.helpmenu.add_cascade(label="Tutorials",
                                  menu=self.tutomenu,
                                  font=self.menu_fontsize)
        self.helpmenu.add_separator()
        self.helpmenu.add_command(label="Open Samples Folder",
                                  command=self.root.open_samples_folder,
                                  font=self.menu_fontsize)
        self.helpmenu.add_command(label="Open Samples Chart App",
                                  command=self.root.open_samples_chart_app,
                                  font=self.menu_fontsize)
        self.add_cascade(label="Help",
                         menu=self.helpmenu,
                         font=self.menu_fontsize)
        # Add to root
        self.visible = visible
        if self.visible:
            master.root.config(menu=self)

    def toggle(self):
        """ Hides/shows this menu """
        self.root.root.config(menu=self if not self.visible else 0)
        self.visible = not self.visible
        return

    def allow_connections(self, **kwargs):
        """
        Starts a new instance of ServerManager.TempoServer and connects it
        with the clock
        """
        if self.listening_for_connections.get() is True:
            Clock = self.root.namespace["Clock"]
            Clock.start_tempo_server(TempoServer, **kwargs)
            print("Listening for connections on {}".format(Clock.tempo_server))
        else:
            Clock = self.root.namespace["Clock"]
            Clock.kill_tempo_server()
            print("Closed connections")
        return

    def start_listening(self, **kwargs):
        """ Manual starting of Renardo tempo server """
        # ToDo - take this method out of the menu
        self.listening_for_connections.set(not self.listening_for_connections.get())
        self.allow_connections(**kwargs)
        return

    def set_cpu_usage(self, *args):
        """ Updates the cpu usage option """
        self.root.namespace["Clock"].set_cpu_usage(self.cpu_usage.get())
        return

    def set_latency(self, *args):
        """ Updates the cpu usage option """
        self.root.namespace["Clock"].set_latency(self.latency.get())
        return


class PopupMenu(Menu):
    def __init__(self, master, **kwargs):
        self.root = master
        Menu.__init__(self, master.root, tearoff=0)
        self.add_command(label="Undo",
                         command=self.root.undo,
                         accelerator="{}+Z".format(ctrl))
        self.add_command(label="Redo",
                         command=self.root.redo,
                         accelerator="{}+Y".format(ctrl))
        self.add_separator()
        self.add_command(label="Copy",
                         command=self.root.edit_copy,
                         accelerator="{}+C".format(ctrl))
        self.add_command(label="Cut",
                         command=self.root.edit_cut,
                         accelerator="{}+X".format(ctrl))
        self.add_command(label="Paste",
                         command=self.root.edit_paste,
                         accelerator="{}+V".format(ctrl))
        self.add_separator()
        self.add_command(label="Select All",
                         command=self.root.select_all,
                         accelerator="{}+A".format(ctrl))
        self.bind("<FocusOut>", self.hide)  # hide when clicked off

    def show(self, event):
        """ Displays the popup menu """
        try:
            self.post(event.x_root, event.y_root)
        finally:
            self.grab_release()

    def hide(self, event):
        """ Removes menu from sight """
        self.unpost()


class ConsolePopupMenu(Menu):
    def __init__(self, master, **kwargs):
        self.root = master
        Menu.__init__(self, master.root, tearoff=0)
        self.add_command(label="Undo",
                         state=DISABLED,
                         accelerator="{}+Z".format(ctrl))
        self.add_command(label="Redo",
                         state=DISABLED,
                         accelerator="{}+Y".format(ctrl))
        self.add_separator()
        self.add_command(label="Copy",
                         command=self.root.edit_copy,
                         accelerator="{}+C".format(ctrl))
        self.add_command(label="Cut",
                         state=DISABLED,
                         accelerator="{}+X".format(ctrl))
        self.add_command(label="Paste",
                         state=DISABLED,
                         accelerator="{}+V".format(ctrl))
        self.add_separator()
        self.add_command(label="Select All",
                         command=self.root.select_all,
                         accelerator="{}+A".format(ctrl))
        self.bind("<FocusOut>", self.hide)  # hide when clicked off

    def show(self, event):
        """ Displays the popup menu """
        try:
            self.post(event.x_root, event.y_root)
        finally:
            self.grab_release()

    def hide(self, event):
        """ Removes menu from sight """
        self.unpost()



#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as f:
    long_description=f.read()

setup(
    name='FoxDotEditor',
    version="0.9.13",
    description='Original FoxDot editor extracted from FoxDot Project - Live coding music with SuperCollider',
    author='Elie Gavoty',
    author_email='eliegavoty@free.fr',
    license='cc-by-sa-4.0',
    url='http://foxdot.org/',
    packages=['FoxDotEditor'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={'gui_scripts': ['FoxDotEditor = FoxDotEditor.__init__:main']},
    # data_files=[('', 'LICENSE')],
    package_data={'FoxDotEditor': ['README.md', 'img/*', 'tmp/*', 'themes/*'], },
    install_requires=[
        'renardo-lib==0.9.13',
        'renardo_gatherer==0.1.4.dev1',
        'psutil',
        'ttkbootstrap==1.10.1',
    ],
)
